/*
* AMT Gateway Win32 version
* 
* File: AMTCtrl.cpp
*
* Description: AMT protocol mangement class
*
* THE UNIVERSITY OF TEXAS AT DALLAS, ANDES Lab.
* Juhyun Oh (Justin), justinoh@student.utdallas.edu
*
*/

#include "AMTCtrl.h"
//#include "Trace.h"

/* decalre critical sedtion globally */
CRITICAL_SECTION CAMTCtrl::g_csReqQ;
CRITICAL_SECTION CAMTCtrl::g_csMsgQ;
//CRITICAL_SECTION CAMTCtrl::g_csFwrdQ;

//CRITICAL_SECTION CAMTCtrl::g_csMsgQ;


/* ==============================================================
*
* Name: CAMTCtrl (Default constructor)
* Parameters:
* Return:
* Descriptions:  initialize required object and variables
*
===============================================================*/
CAMTCtrl::CAMTCtrl(void)
{
	/* initialize randm */
	srand( (unsigned)time( NULL ) );

	/* critical section initialization */
	InitializeCriticalSection( &g_csReqQ );
	InitializeCriticalSection( &g_csMsgQ );
	//InitializeCriticalSection( &g_csFwrdQ );

	/* */
	ReadConfigFile();

	/* initialze socket and ... */
	InitializeTimerQueue();
	InitializeSocket();
	InitializeAMTProtocolHandler();
	InitializeReceiver();

	
	/* relay status */
	SetRelayStatus(RELAY_NOT_FOUND);
	//SetRelayStatus(RELAY_FOUND);

	/* init discovery timer handle  */
	m_hDisoveryTimer  = NULL;
	m_hIGMPQueryTimer = NULL;
}

CAMTCtrl::~CAMTCtrl(void)
{
	/* clean up */

	/*  sockets */
	closesocket( m_sUDP );
	closesocket( m_sDiscovery );
	//closesocket( m_sSend );

	/* timer Q handle */
	if( NULL != m_hTimerQueue)
	{
		DeleteTimerQueueEx(m_hTimerQueue,NULL);
		//CloseHandle( m_hTimerQueue );
		m_hTimerQueue = NULL;
	}

	/* receive notification event */
	if( NULL != m_hRcvEvent )
	{
		CloseHandle( m_hRcvEvent );
		m_hRcvEvent = NULL;
	}


	/* receive queue - deallocate memory */
	if( m_listMsgQ.size() > 0 )
	{
		list<RELAY_MSG>::iterator  iter;
		//char*                    pMsgBuffer;
		RELAY_MSG                  RelayMsg;

		/* deallocate the memory */
		for( iter = m_listMsgQ.begin(); iter != m_listMsgQ.end(); iter++ )
		{
			RelayMsg = (*iter);
			if( NULL != RelayMsg.pcData )
			{
				delete [] RelayMsg.pcData;
			}
		}
		/* remove all item in the queue */
		m_listMsgQ.clear();
	}


	/* check pending queue and deallocate memory if something left */
	if( m_listReqestQ.size() > 0 )
	{
#if 1
		list<REQUEST>::iterator    iter;
		REQUEST                    request;
		for( iter = m_listReqestQ.begin(); iter != m_listReqestQ.end(); iter++ )
		{
			request = *iter;
			if( NULL != request.pucIGMPData )
			{
				///printf("\n[-]IGMP_REQ: 0x%x",request.pucIGMPData);
				delete [] request.pucIGMPData;
			}
		}
#endif

		/* remove items */
		m_listReqestQ.clear();
	}

	/* Delete Critical section */
	DeleteCriticalSection( &g_csReqQ );
	DeleteCriticalSection( &g_csMsgQ );
	//DeleteCriticalSection( &g_csFwrdQ );
}


void CAMTCtrl::InitializeSocket(void)
{
	int nRet = 0;
	SOCKADDR_IN rcvAddr;
	SOCKADDR_IN rcvAddr2;

	ZeroMemory( &rcvAddr, sizeof(SOCKADDR_IN) );


	/* create discovery socket */
	m_sDiscovery = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP );

	if( INVALID_SOCKET == m_sDiscovery )
	{
		/* error */
		DbgPrint("    [AMTCtrl] Fail to create discovery socket (err:%d)", WSAGetLastError() ) ;
		::MessageBox(NULL, TEXT("disco socket creation fail"), NULL, MB_OK | MB_ICONEXCLAMATION );
		return;
	}

	/* Fill in the address structure */
	rcvAddr.sin_family      = AF_INET;
	rcvAddr.sin_port        = htons(0);   
	rcvAddr.sin_addr.s_addr = INADDR_ANY; // Let WinSock assign active interface (address)

	
	/* bind the name to the socket */
	nRet = bind(m_sDiscovery,	                // Socket descriptor
				(LPSOCKADDR)&rcvAddr,   // Address to bind to
				sizeof(struct sockaddr)	// Size of address
				);
	
	if (SOCKET_ERROR == nRet)
	{
		DbgPrint("    [AMTCtrl] error at discovery socket binding (err: %d)", WSAGetLastError() );
		closesocket(m_sDiscovery);
		::MessageBox(NULL, TEXT("disco socket binding fail"), NULL, MB_OK | MB_ICONEXCLAMATION );
		return;
	}
	
	/* fill in relay anycast address for discovery */
	m_addrDiscovery.sin_family      = AF_INET;
	m_addrDiscovery.sin_port        = htons(AMT_PORT);
	m_addrDiscovery.sin_addr.s_addr = inet_addr(m_strRelayIP.c_str()); 


	/* create UDP socket */
	m_sUDP = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP );

	if( INVALID_SOCKET == m_sUDP )
	{
		/* error */
		DbgPrint("    [AMTCtrl] Fail to create UDP soecket", WSAGetLastError() );
		::MessageBox(NULL, TEXT("msg socket creation fail"), NULL, MB_OK | MB_ICONEXCLAMATION );
		return;
	}

	/* Fill in the address structure */
	ZeroMemory(&rcvAddr2, sizeof(rcvAddr2) );
	rcvAddr2.sin_family      = AF_INET;
	rcvAddr2.sin_port        = htons(0);   
	rcvAddr2.sin_addr.s_addr = INADDR_ANY; // Let WinSock assign active interface (address)

	/* binding */
	nRet = bind( m_sUDP, 
		         (LPSOCKADDR)&rcvAddr2, 
				 sizeof(struct sockaddr) );
	
	if( SOCKET_ERROR == nRet )
	{
		//printf("\n    [AMTCtrl] Failt to bind UDP soecket", WSAGetLastError() );
		DbgPrint("    [AMTCtrl] Fail to bind UDP soecket", WSAGetLastError() );
		closesocket(m_sUDP);
		::MessageBox(NULL, TEXT("msg socket binding fail"), NULL, MB_OK | MB_ICONEXCLAMATION );
		return;
	}

	m_sQuery = socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP );
	if( INVALID_SOCKET == m_sQuery )
	{
		/* error */
		DbgPrint("    [AMTCtrl] Fail to create query socket (err:%d)", WSAGetLastError() ) ;
		return;
	}

	// binding socket to local address
	m_stLocalAddr.sin_family      = AF_INET;
	m_stLocalAddr.sin_port        = htons(0);   
	m_stLocalAddr.sin_addr.s_addr = INADDR_ANY;

	nRet = bind(
			m_sQuery,					// Socket descriptor
			(sockaddr*)&m_stLocalAddr,	// Address to bind to
			sizeof(struct sockaddr)	);	// Size of address
			

	// server aaddress
	m_stSvrAddr.sin_family      = AF_INET;
	m_stSvrAddr.sin_port        = htons(SVR_PORT);
	m_stSvrAddr.sin_addr.s_addr = inet_addr("127.0.0.1");


	/* make discovery socket to non-blocking socket */
	// do we really required?
}


void CAMTCtrl::InitializeReceiver(void)
{
	/* create thread and check returned handle */
	m_hDiscoveryRcvThread = (HANDLE)_beginthread( DiscoveryRcvProc, 0, (void*)this );
	
	if( -1L ==  (long)m_hDiscoveryRcvThread )
	{
		DbgPrint("    [AMTCtrl] Failt to create disco. msg receiver thread");
	}

#if 1
	m_hRcvThread = (HANDLE)_beginthread( MsgRcvProc, 0, (void*)this );
	
	if( -1L ==  (long)m_hRcvThread )
	{
		DbgPrint("    [AMTCtrl] Failt to create UDP msg receiver thread");
	}
	
#endif
	
}

void CAMTCtrl::InitializeAMTProtocolHandler(void)
{
	/* create thread and check returned handle */
	m_hAMTHandlerThread = (HANDLE)_beginthread( AMTProtocolHandler, 0, (void*)this );

	if( -1L ==  (long)m_hAMTHandlerThread )
	{
		DbgPrint("    [AMTCtrl] AMT handler thread creation failed" );
	}

	/* create event for reception notifications */
	m_hRcvEvent = ::CreateEvent( NULL,                    // security attribute (default)
		                         FALSE,                    // reset type (manual reset)
				                 FALSE,                   // initial state (not signaled)
				                 TEXT("_MY_RECP_SIGNAL_")    // event name
				                );
	if( NULL == m_hRcvEvent )
	{
		DbgPrint("    [AMTCtrl] Fail to create receive event" );
	}

}

void CAMTCtrl::InitializeTimerQueue( void )
{
	/* create timer queue handle - use kernel provide timer pool */
	m_hTimerQueue = CreateTimerQueue();

	if (NULL == m_hTimerQueue)
    {
		DbgPrint("    [AMTCtrl] Failt to create timer queue (err: %d)", GetLastError() );
    }
}

#if 0
int CAMTCtrl::GetRandom(void)
{
	return rand();
}
#endif


/* let's do it quick and dirty right now...*/
BOOL CAMTCtrl::ReadConfigFile( void )
{

	ifstream    inFile;
	char        chReadBuffer[1024];
	string      strTemp;
	
//#if defined(_DEBUG)
//	inFile.open("..\\AMTGateway.conf", ifstream::in);
//#else
	inFile.open("AMTGateway.conf", ifstream::in);
//#endif

	if( !inFile.good() )
	{
		//printf("\n    [AMTCtrl] Fail to open AMTGateway.conf file. Set Default IP: 129.110.95.65");
		DbgPrint("[AMTCtrl] Fail to open AMTGateway.conf file. Set Default relay IP: 154.17.0.1");
		m_strRelayIP = "154.17.0.1";
		//m_strRelayIP = "129.110.95.61";
		return FALSE;
	}
	
	while( inFile.good() )
	{
		ZeroMemory( chReadBuffer, sizeof(chReadBuffer) );
		inFile.getline(chReadBuffer, sizeof(chReadBuffer) );
		
		strTemp = chReadBuffer;
		
		if( ';' == strTemp[0] )
		{
			/* skip comment */
			continue;
		}

		if( strTemp.size() > 0  )
		{
			m_strRelayIP = strTemp;
			//printf("\n[AMTCtrl] Set realy IP: %s", m_strRelayIP.c_str() );
			DbgPrint("[AMTCtrl] Set realy IP: %s", m_strRelayIP.c_str() );
			break;
		}
	}

	return TRUE;

}



void CAMTCtrl::CreateDiscoveryTimer( void )
{
	BOOL bSuccess;

	/* initialization */
	bSuccess = FALSE;

	/* check current discovery state and create one if there is none */
	if( NULL == m_hDisoveryTimer && RELAY_NOT_FOUND == m_nRelayStatus )
	{
		/* update relay status */
		SetRelayStatus(RELAY_DISCOVERY_INPROGRESS);

		/* the timer expieres immediately */
		bSuccess = CreateTimerQueueTimer( &m_hDisoveryTimer,       // handle to timer
										  m_hTimerQueue,           // handle to timer queue
										  (WAITORTIMERCALLBACK)DiscoveyTimerHandler, // timer callback function
										  this,                    // callback parameter
										  0,                       // timer due time
										  DISCOVERY_TIMER_PERIOD,  // timer period
										  WT_EXECUTEINTIMERTHREAD  // options
										  );
	}
	else
	{
		if( m_hDisoveryTimer != NULL )
		{
			SetRelayStatus(RELAY_DISCOVERY_INPROGRESS);
		}
		DbgPrint("[AMTCtrl] Disco Timer Hndl = %d, Relay Status = %d", m_hDisoveryTimer, m_nRelayStatus );
	}

}

void CAMTCtrl::DeleteDiscoveryTimer( void )
{
	BOOL bSuccess;
	
	/* initializations */
	bSuccess = FALSE;

	/*delete timer */
	bSuccess = DeleteTimerQueueTimer( m_hTimerQueue,     // handle to timer queue
                                      m_hDisoveryTimer,  // handle to timer
                                      NULL               // handle to completion event
									  );
	if( FALSE == bSuccess )
	{
		int nError = GetLastError();
		if( ERROR_IO_PENDING != nError )
		{
			/* error */
			DbgPrint("    [AMTCtrl] Fail to remove discovery timer (%d)", nError );
		}
	}
	else
	{
		/* reset to initial state */
		DbgPrint("[AMTCtrl] Remove Discovery timer");
		m_hDisoveryTimer = NULL;
	}

}

void CAMTCtrl::DiscoveyTimerHandler( PVOID lpParameter, BOOLEAN TimerOrWaitFired )
{
	CAMTCtrl* pParent;

	//srand( (unsigned)time( NULL ) );


	pParent = (CAMTCtrl*)lpParameter;

	//printf("\n disco nonce: %d", pParent->GetRandom() );


	/* send discovery */
	pParent->SendRelayDiscovery();
}

/*
*  CreateIGMPQueryTimer( void )
*
*/

void CAMTCtrl::CreateIGMPQueryTimer( void )
{
	BOOL bSuccess = FALSE;

	/* check current discovery state and create one if there is none */
	if( NULL == m_hIGMPQueryTimer )
	{
		/* the timer expieres immediately */
		bSuccess = CreateTimerQueueTimer( &m_hIGMPQueryTimer,							// handle to timer
										  m_hTimerQueue,								// handle to timer queue
										  (WAITORTIMERCALLBACK)IGMPQueryTimerHandler,	// timer callback function
										  this,											// callback parameter
										  60000,										// timer due time
										  IGMPQUERY_TIMER_PERIOD,						// timer period
										  WT_EXECUTELONGFUNCTION );						// options
										  
	}
	
}

void CAMTCtrl::DeleteIGMPQueryTimer( void )
{
	BOOL bSuccess = FALSE;

	/*delete timer */
	bSuccess = DeleteTimerQueueTimer( m_hTimerQueue,		// handle to timer queue
                                      m_hIGMPQueryTimer,	// handle to timer
                                      NULL );				// handle to completion event
	
	/* always false because this func called from the handler */								  
	if( FALSE == bSuccess )
	{
		int nError = GetLastError();
		if( ERROR_IO_PENDING != nError )
		{
			/* error */
			DbgPrint("    [AMTCtrl] Fail to remove IGMP query timer (%d)", nError );
		}
		else
		{
			/* reset to initial state */
			DbgPrint("[AMTCtrl] Remove IGMP query timer");
			m_hIGMPQueryTimer = NULL;
		}
	}
	else
	{
		/* reset to initial state */
		DbgPrint("[AMTCtrl] Remove IGMP query timer");
		m_hIGMPQueryTimer = NULL;
	}

}

/*
unsigned short CAMTCtrl::GetChecksum(unsigned short *pusIPBuffer, int nLen)
{
    int nleft = nLen;
    int sum = 0;
    unsigned short *w = pusIPBuffer;
    unsigned short answer = 0;

    while (nleft > 1) 
	{
        sum += *w++;
        nleft -= 2;
    }
    if (nleft == 1) 
	{
        *(unsigned char*)(&answer) = *(unsigned char*)w;
        sum += answer;
    }
    sum = (sum >> 16) + (sum & 0xffff);
    answer = ~sum;
    return (answer);
}


void CAMTCtrl::SetIFAddress( string strIP )
{
	m_strNicIP = strIP;
}
*/

// IGMP query timer handler
void CAMTCtrl::IGMPQueryTimerHandler( PVOID lpParameter, BOOLEAN TimerOrWaitFired )
{
	CAMTCtrl* pParent = (CAMTCtrl*)lpParameter;

	char pBuffer[1];	
	pBuffer[0] = PERIODIC_QUERY;
	
	int nRet = sendto( 
				pParent->m_sQuery, 
				pBuffer, 
				1, 
				0, 
				(sockaddr*)&pParent->m_stSvrAddr, 
				sizeof(struct sockaddr) );
}




void CAMTCtrl::SetRelayStatus( int nStatus )
{
	m_nRelayStatus = nStatus;
	
	switch(nStatus)
	{
		case RELAY_NOT_FOUND:
			//printf("\n[AMTCtrl] Relay ststus update: NOT FOUND");
			DbgPrint("[AMTCtrl] Relay status update: NOT FOUND");
			break;
		case RELAY_FOUND:
			//printf("\n[AMTCtrl] Relay ststus update: FOUND");
			DbgPrint("[AMTCtrl] Relay status update: FOUND");
			break;
		case RELAY_DISCOVERY_INPROGRESS:
			//printf("\n[AMTCtrl] Relay ststus update: DISOCVERY");
			DbgPrint("[AMTCtrl] Relay status update: DISOCVERY IN-PROGRESS");
			break;
	}
	
}

int CAMTCtrl::GetRelayStstus( void )
{
	return m_nRelayStatus;
}



void CAMTCtrl::SetPktSenderObject( CPktSender* pObject )
{
	m_pObjPktSender = pObject;
}

#if 0
void CAMTCtrl::SetGroupInfoObject( CCurrentGroupInfo* pObject )
{
	m_pObjGroupInfo = pObject;
}
#endif



void CAMTCtrl::OnAMTAdvertisementMsg( char* pBuffer )
{
	int nRet;
	unsigned int ulMyNonce;
	unsigned int ulRcvedNonce;
	unsigned long ulNonce;
	unsigned long ulRelayIP;
	//SOCKADDR_IN addrLocal;
	

	/* initialization */
	nRet         = 0;
	ulMyNonce    = 0;
	ulRcvedNonce = 0;
	ulNonce      = 0;
	//ZeroMemory( &addrLocal, sizeof(SOCKADDR_IN) );
	
	/* are we looking for relay? */
	if( RELAY_DISCOVERY_INPROGRESS == GetRelayStstus() && 
		NULL                       != m_hDisoveryTimer )
	{
		/* get nonce from received message */
		memcpy(&ulNonce, &pBuffer[4], RESPONSE_NONCE_LEN );

		/* conver it to host order */
		ulRcvedNonce = ntohl(ulNonce);
		//printf("\n[AMTCtrl] Rcvd relay nonce : %d", ulRcvedNonce);
		//printf("\n[AMTCtrl] Local relay nonce: %d", m_ulRelayNonce);
		DbgPrint("[AMTCtrl] Rcvd relay nonce : %d, 0x%x", ulRcvedNonce, ulRcvedNonce);
		DbgPrint("[AMTCtrl] Local relay nonce: %d, 0x%x", m_ulRelayNonce, ulRcvedNonce);

		/* check curernt nonce and received nonce */
		if( m_ulRelayNonce == ulRcvedNonce )
		{
			/* copy relay IP from payload */
			memcpy( &ulRelayIP, &pBuffer[8], RESPONSE_RELAYIP_LEN );
			//ulRelayIP = ntohl(ulRelayIP);
			//printf("\n[AMTCtrl] ulRelayIP: %x", ulRelayIP);

			/* copy relay address for future use */
			ZeroMemory(&m_addrRelay, sizeof(SOCKADDR_IN) );
			
			m_addrRelay.sin_family        = AF_INET;
			//bug // m_addrRelay.sin_addr.s_addr   = m_addrSender.sin_addr.s_addr; 
			//m_addrRelay.sin_addr.s_addr   = htonl(ulRelayIP);
			m_addrRelay.sin_addr.s_addr   = ulRelayIP;
			m_addrRelay.sin_port          = htons(AMT_PORT);

			//printf("\n[AMTCtrl] Received relay IP address: %s", inet_ntoa(m_addrRelay.sin_addr) );
			DbgPrint("[AMTCtrl] Received relay IP address: %s", inet_ntoa(m_addrRelay.sin_addr) );
			

			/* update relay status */
			SetRelayStatus(RELAY_FOUND);

			/* kill discovery timer */
			DeleteDiscoveryTimer();

			/* original code - make connection to the relay 
			 * -> notify when relay down & use send() rather than sendto()
			 */
#if 1
			/* connection */
			nRet = connect( m_sUDP, (SOCKADDR*)&m_addrRelay, sizeof(m_addrRelay) );
			
			if( SOCKET_ERROR == nRet )
			{
				//printf("\n    [AMTCtrl] UDP socket connection error (err: %d)", WSAGetLastError() );
				DbgPrint("    [AMTCtrl] UDP socket connection error (err: %d)", WSAGetLastError() );
			}
#endif

			/* record current time */
			//time(&m_ulTunnelUpTime);
			
			/* check pending request and let them be serviced */
			ProcessRequest();

		}
		else
		{
			/* just ignore? */
			DbgPrint("    [AMTCtrl] received advertisement but nonce dismatch - ignore");
		}
	}
	else
	{
		DbgPrint("    [AMTCtrl] Not in discovery but received advertisement - ignore");
	}
}

void CAMTCtrl::OnAMTMembershipQueryMsg( char* pBuffer )
{
	unsigned long    ulNonce;
	unsigned long    ulRcvedNonce;
	unsigned char    uchaMAC[6];
	list<REQUEST>::iterator iter;

	unsigned char*   pIGMPPkt;
	unsigned int     nPktLength;
	BOOL             bRet;
	BOOL             bFound;

	/* initializations */
	pIGMPPkt   = NULL;
	nPktLength = 0;
	
	/* check message length */

	/* check MAC length */

	/*
	 *  Message format 
	 * +-------------------------------------------------------------------+
	 * | Msg Type(1 byte)| Reserved (1 byte)| MAC (6 byte)| nonce (4 byte) |
	 * +-------------------------------------------------------------------+ 
	 */

	/* copy MAC */
	memcpy(&uchaMAC[0], &pBuffer[2], RESPONSE_MAC_LEN );

	/* copy nonce */
	memcpy( &ulRcvedNonce, &pBuffer[8], RESPONSE_NONCE_LEN );
	ulNonce = ntohl( ulRcvedNonce );

	//printf("\n[AMTCtrl] Rcved membership query from relay (nonce: %d, MAC: 0x%x-0x%x-0x%x-0x%x-0x%x-0x%x)", ulNonce, 
	//	    (unsigned char)chaMAC[0], (unsigned char)chaMAC[1], (unsigned char)chaMAC[2], 
	//		(unsigned char)chaMAC[3], (unsigned char)chaMAC[4], (unsigned char)chaMAC[5]);
	DbgPrint("[AMTCtrl] Rcved membership query from relay (nonce: %d, 0x%x,MAC: 0x%x:0x%x:0x%x:0x%x:0x%x:0x%x)", ulNonce, ulNonce,
		    (unsigned char)uchaMAC[0], (unsigned char)uchaMAC[1], (unsigned char)uchaMAC[2], 
			(unsigned char)uchaMAC[3], (unsigned char)uchaMAC[4], (unsigned char)uchaMAC[5]);


	/* find nonce in the request queue */
	EnterCriticalSection(&g_csReqQ);
	bFound= FALSE;
	for( iter = m_listReqestQ.begin(); iter != m_listReqestQ.end(); iter++ )
	{
		if( iter->ulNonce == ulNonce )
		{
			/* find right one - stop searching */
			//printf("\n[AMTCtrl] Found match (nonce: %d)", ulNonce );
			DbgPrint("[AMTCtrl] Found match (nonce: %d, 0x%x)", ulNonce, ulNonce );
			bFound = TRUE;
			break;
		}
	}
	//LeaveCriticalSection( &g_csReqQ );

	if( TRUE == bFound )
	{
		pIGMPPkt   = new unsigned char[iter->ulIGMPLength];
		nPktLength = iter->ulIGMPLength;

		if( NULL == pIGMPPkt )
		{
			//printf("\n    [AMTCtrl] temp buffer creation fail");
			DbgPrint("    [AMTCtrl] temp buffer creation fail");
		}
		else
		{
#ifdef MEM_DEBUG
			char szMessage[1024];
			ZeroMemory( szMessage, sizeof(szMessage) );
			int nSize = sprintf_s(szMessage, "0x%08x@File: %s, Line:%d", pIGMPPkt, __FILE__, __LINE__);
			Mem_Debug(szMessage, nSize, TRUE );
#endif
		}


		memcpy( pIGMPPkt, iter->pucIGMPData, iter->ulIGMPLength);

		/* increase retry counter to maximum+1 */
		iter->ulReqCount = GW_REQUEST_MAX+1;

		/* adjust timer interval so that the request can be deleted after expiration time */
		int nDue = GW_REQUEST_RETRY_INTERVAL * 3 + 2000; // 5 sec

		bRet = ChangeTimerQueueTimer( 
							m_hTimerQueue, 
							iter->hReqTimerHndl, 
							nDue,
							REQ_TIMER_INIT_PERIOD );
		
		DbgPrint("[AMTCtrl] expire after: (+)%d ms", nDue);
		if( FALSE == bRet )
		{
			DbgPrint("    [AMTCtrl] fail to change request timer" );
		}

#if 0
		/* remove request from the request queue */
		DbgPrint("[AMTCtrl] Remove request (nonce: %d, 0x%x)",ulNonce, ulNonce);

		/* delete timer */
		bRet = DeleteTimerQueueTimer( m_hTimerQueue, iter->hReqTimerHndl, NULL);
		if( FALSE == bRet )
		{
			int nError = GetLastError();
			if( ERROR_IO_PENDING != nError )
			{
				DbgPrint("    [AMTCtrl] Fail to remove req. timer (err: %d)", nError );
			}
		}
		
		/* dealloc ReqInfo and IGMP buffer */
		if( NULL != iter->pReqInfo )
		{
			///printf("\n[-]REQ_INFO:0x%x", iter->pReqInfo );
#ifdef MEM_DEBUG
			char szMessage[1024];
			ZeroMemory( szMessage, sizeof(szMessage) );
			int nSize = sprintf_s(szMessage, "0x%08x@File: %s, Line:%d", iter->pReqInfo, __FILE__, __LINE__);
			Mem_Debug(szMessage, nSize, FALSE );
#endif
			delete iter->pReqInfo;
		}

		if( NULL != iter->pucIGMPData )
		{
			///printf("\n[-]IGMP_REQ: 0x%x",iter->pucIGMPData);
#ifdef MEM_DEBUG
			char szMessage[1024];
			ZeroMemory( szMessage, sizeof(szMessage) );
			int nSize = sprintf_s(szMessage, "0x%08x@File: %s, Line:%d", iter->pucIGMPData, __FILE__, __LINE__);
			Mem_Debug(szMessage, nSize, FALSE );
#endif
			delete [] iter->pucIGMPData;
		}

		//EnterCriticalSection(&g_csReqQ);
		iter = m_listReqestQ.erase(iter);
		//LeaveCriticalSection( &g_csReqQ );
#endif
	}
	else
	{
		//printf("\n[AMTCtrl] could NOT find nonce in the request list");
		//printf("\n[AMTCtrl] ignore membership query  (nonce:%d)", ulNonce);
		DbgPrint("[AMTCtrl] could NOT find nonce in the request list");
		DbgPrint("[AMTCtrl] ignore membership query  (nonce:%d, 0x%x)", ulNonce, ulNonce);
	}
		
	LeaveCriticalSection( &g_csReqQ );

	/* send query to kernel - just once */


	if( TRUE == bFound )
	{
		/* send memebership change */
		SendMembership( &uchaMAC[0], ulNonce, pIGMPPkt, nPktLength );
	
		/* deallocate temp memory */
		if( NULL != pIGMPPkt )
		{
#ifdef MEM_DEBUG
			char szMessage[1024];
			ZeroMemory( szMessage, sizeof(szMessage) );
			int nSize = sprintf_s(szMessage, "0x%08x@File: %s, Line:%d", pIGMPPkt, __FILE__, __LINE__);
			Mem_Debug(szMessage, nSize, FALSE );
#endif
			delete [] pIGMPPkt;	
		}
	}
	

}


void CAMTCtrl::AMTProtocolHandler( void* pParam )
{
	int        nRet;         // check return
	CAMTCtrl*  pParent;
	int        nMsgType;
	//HANDLE     hReceiveSignal;
	DWORD      dwRet;
	//char*      pRcvBuffer;
	RELAY_MSG  RelayMsg;

	/* initialization */
	nRet     = 0;
	nMsgType = 0;
	pParent  = (CAMTCtrl*)pParam;

	//hReceiveSignal = pParent->m_hRcvEvent; 
	int nMsgQSize = 0;
	int nFowardQSize = 0;
	//int test =0;

	/* initialize randm */
	//srand( (unsigned)time( NULL ) );

	while(1) 
	{
		/* wait until a sginal comes */

		//printf( "\n(%d, %d)", nMsgQSize, nFowardQSize );
		if( (0 == nMsgQSize) && (0 == nFowardQSize) )
		{
			/* wait until a sginal comes */
			dwRet = ::WaitForSingleObject( pParent->m_hRcvEvent, INFINITE );

			EnterCriticalSection(&g_csMsgQ);
			//EnterCriticalSection(&g_csFwrdQ);

			nMsgQSize    = pParent->m_listMsgQ.size();
			nFowardQSize = pParent->m_listForwardQ.size();
		
			//LeaveCriticalSection(&g_csFwrdQ);
			LeaveCriticalSection(&g_csMsgQ);

		}

#if 1
		/* Mcast forwarding has lower priority */
		if( nMsgQSize > 0 )
		{
			/* retrieve and remove */
			EnterCriticalSection(&g_csMsgQ);
			
			//test = pParent->m_listMsgQ.size();
			//if( pParent->m_listMsgQ.size() > 0 )
			{
				RelayMsg = pParent->m_listMsgQ.front();
				pParent->m_listMsgQ.pop_front();
			}

			LeaveCriticalSection(&g_csMsgQ);
		}
		else if( nFowardQSize > 0 )
		{
			/* retrieve and remove */
			EnterCriticalSection(&g_csMsgQ);
			//test = pParent->m_listForwardQ.size();
			//if( pParent->m_listForwardQ.size() > 0 )
			{
				RelayMsg = pParent->m_listForwardQ.front();
				pParent->m_listForwardQ.pop_front();
			}

			LeaveCriticalSection(&g_csMsgQ);
		}
		else
		{
			
			EnterCriticalSection(&g_csMsgQ);
			nMsgQSize    = pParent->m_listMsgQ.size();
			nFowardQSize = pParent->m_listForwardQ.size();
			LeaveCriticalSection(&g_csMsgQ);			
			
			DbgPrint("[AMTCtrl] !!! No data but signaled !!! %d, %d",nMsgQSize, nFowardQSize);
			
			if( ::ResetEvent(pParent->m_hRcvEvent) == 0 )
			{
				DbgPrint("    [AMTCtrl](1) Fail reset event (%d)", GetLastError() );
			}
			continue;
		}
#endif		


		/* parsing received mesasge and do something according to the AMT protocol */
		nMsgType = (int)RelayMsg.pcData[0];	
		
		switch( nMsgType )
		{
			case AMT_RELAY_DISCOVERY:
				/* GW sendinge message - nothing to do */
				break;

			case AMT_RELAY_ADVERTISEMENT:
				/* respone of the discovery message */			
				DbgPrint("[AMTCtrl] Relay->GW (ADVERTISEMENT)");
				pParent->OnAMTAdvertisementMsg(RelayMsg.pcData);
				break;

			case AMT_REQUEST:
				/* GW sendinge message - nothing to do */
				break;

			case AMT_MEMBERSHIP_QUERY:
				/* reponse of the request */
				DbgPrint("[AMTCtrl] Relay->GW (MEMBERSHIP QUERY)");
				pParent->CreateIGMPQueryTimer();
				pParent->OnAMTMembershipQueryMsg(RelayMsg.pcData);
				break;

			case AMT_MEMBERSHIP_CHANGE:
				/* nothing to do */
				break;

			case AMT_MCAST_DATA:
				/* pass the packet to kernel */
				pParent->ForwardingMcastData( RelayMsg.pcData, RelayMsg.nDataLength );
				break;

			default:
				DbgPrint("    [AMTCtrl]Invalid message from relay");
				break;
	
		}// end of switch satement ////////////////////

		/* deallocate memory */
		///printf("\n[-] RELAY_MSG: 0x%x", RelayMsg.pcData );
		
#ifdef MEM_DEBUG
		char szMessage[1024];
		ZeroMemory( szMessage, sizeof(szMessage) );
		int nSize = sprintf_s(szMessage, "0x%08x@File: %s, Line:%d", RelayMsg.pcData, __FILE__, __LINE__);
		Mem_Debug(szMessage, nSize, FALSE );
#endif
		delete [] RelayMsg.pcData;

		EnterCriticalSection(&g_csMsgQ);

		nMsgQSize    = pParent->m_listMsgQ.size();
		nFowardQSize = pParent->m_listForwardQ.size();

		LeaveCriticalSection(&g_csMsgQ);

        if ( ::ResetEvent(pParent->m_hRcvEvent) == 0 )
		{
			DbgPrint("    [AMTCtrl](2) Fail reset event (%d)", GetLastError() );
		}


	}// end of while loop /////////////////////

	/* don't forget to free receiver buffer */

}




void CAMTCtrl::DiscoveryRcvProc(void* pParam)
{
	CAMTCtrl*      pParent;      // this pointer
	int            nRet;
	unsigned long  ulByteRcved;  // for verifying receiving data size
	char*          pRcvBuffer;   // receiver buffer 
	int            nBufferSize;
	int            nAddrSize;    // size of sockaddr_in
	RELAY_MSG      RelayMsg; 
	BOOL           bRet;
	fd_set         readFDs;	     // read fd set

	/* initialization */
	pParent     = (CAMTCtrl*)pParam;
	nRet        = 0;
	ulByteRcved = 0;
	pRcvBuffer  = NULL;
	nBufferSize = 0;
	bRet        = FALSE;
	
	
	/* initialize fd set */
	FD_ZERO(&readFDs);
	FD_SET( pParent->m_sDiscovery, &readFDs );

	while(1)
	{
		/* wait here until get some messages */
		nRet = select( FD_SETSIZE, &readFDs, NULL, NULL, NULL );
		//nRet = select( FD_SETSIZE, &readFDs, NULL, &readFDs, NULL );

		if( SOCKET_ERROR == nRet  )
		{
			DbgPrint("    [AMTCtrl] select() error (%d)", WSAGetLastError() );
		}

		/* check size of receiving buffer */
		nRet = ioctlsocket( pParent->m_sDiscovery, FIONREAD, &ulByteRcved );

		/* allocate memory as much as received size and receive it */
		RelayMsg.pcData = new char[ulByteRcved];


		if( NULL == RelayMsg.pcData )
		{

			/* system may out of memory - what shall we do?? */
			DbgPrint("    [AMTCtrl] DiscoveryRcvProc:Memory allocation failed");
			continue;
		}
		else
		{
#ifdef MEM_DEBUG
			char szMessage[1024];
			ZeroMemory( szMessage, sizeof(szMessage) );
			int nSize = sprintf_s(szMessage, "0x%08x@File: %s, Line:%d", RelayMsg.pcData, __FILE__, __LINE__);
			Mem_Debug(szMessage, nSize, TRUE );
#endif
		}


		/* receive the data from network queue */
		//nBufferSize = ulByteRcved;
		RelayMsg.nDataLength = ulByteRcved;

		
		nAddrSize   = sizeof(SOCKADDR_IN);
		ZeroMemory(&pParent->m_addrSender, sizeof(SOCKADDR_IN) );

		nRet = recvfrom(pParent->m_sDiscovery, 
						RelayMsg.pcData, //pRcvBuffer, 
						RelayMsg.nDataLength, //nBufferSize, 
						0, 
						(SOCKADDR *)&pParent->m_addrSender, 
						&nAddrSize);
	
		if( SOCKET_ERROR  == nRet )
		{
			if( WSAECONNRESET == WSAGetLastError() )
			{
				/* do something properly */
				DbgPrint("    [AMTCtrl] No connection");
#ifdef MEM_DEBUG
				char szMessage[1024];
				ZeroMemory( szMessage, sizeof(szMessage) );
				int nSize = sprintf_s(szMessage, "0x%08x@File: %s, Line:%d", RelayMsg.pcData, __FILE__, __LINE__);
				Mem_Debug(szMessage, nSize, FALSE );
#endif			
				delete [] RelayMsg.pcData;
	
				continue;
			}
			else
			{
				/* we may need to retry based on error code */
				DbgPrint("    [AMTCtrl] receive error (%d)\n", WSAGetLastError()  );
			}
		}
		
		/* put read data into our queue for processing */
		EnterCriticalSection(&g_csMsgQ);
		//printf("\n[+] RELAY_MSG: 0x%x", RelayMsg.pcData );
		pParent->m_listMsgQ.push_back(RelayMsg);
		
		LeaveCriticalSection(&g_csMsgQ);

		bRet = ::SetEvent(pParent->m_hRcvEvent);

		if( FALSE == bRet )
		{
			/* event signaling error */
			//printf("\n    [AMTCtrl] SetEvent error (err: %d)", GetLastError() );
			DbgPrint("    [AMTCtrl] SetEvent error (err: %d)", GetLastError() );
		}
		
	}// end of while loop ////////////////////////////////
}

#if 0
void CAMTCtrl::MsgRcvProc(void* pParam)
{

	int            nRet;
	unsigned long  ulByteRcved;
	char*          pRcvBuffer;   // receiver buffer 
	int            nBufferSize;
	int            nAddrSize;    // size of sockaddr_in
	CAMTCtrl*      pParent; 
	BOOL           bRet;
	SOCKET         sRcv;
	

	/* initialization */
	pParent     = (CAMTCtrl*)pParam;
	nRet        = 0;
	ulByteRcved = 0;
	pRcvBuffer  = NULL;
	nBufferSize = 0;
	bRet        = FALSE;

    nAddrSize   = sizeof(SOCKADDR_IN);

	/* register receiver socket to select function */
	fd_set readFDs;

	/* initialize fd set */
	FD_ZERO(&readFDs);
	FD_SET( pParent->m_sDiscovery, &readFDs );
	FD_SET( pParent->m_sUDP, &readFDs );

	int intface;

	while(1)
	{
		/* wait here until get some messages */
		//nRet = select( FD_SETSIZE, &readFDs, NULL, NULL, NULL );
		nRet = select( 2, &readFDs, NULL, &readFDs, NULL );

		if( SOCKET_ERROR == nRet  )
		{
			//cout << "Select error (" << WSAGetLastError() << ")" << endl;
			printf("Select error (%d)\n", WSAGetLastError() );
		}

		/* check size of receiving buffer */
		nRet = ioctlsocket( pParent->m_sUDP, FIONREAD, &ulByteRcved );

		if( ulByteRcved > 0 )
		{
			sRcv = pParent->m_sUDP;

			intface = 1;

		}
		else
		{
			nRet = ioctlsocket( pParent->m_sDiscovery, FIONREAD, &ulByteRcved );
			if( ulByteRcved > 0 )
			{
				sRcv = pParent->m_sDiscovery;
				intface = 2;
			}
		}

		/* allocate memory as much as received size and receive it */
		pRcvBuffer  = new char[ulByteRcved];
		if( NULL == pRcvBuffer )
		{
			/* system may out of memory - what shall we do?? */
			cout<< "Memory allocation failed" << endl;
			return;
		}

		/* receive the data from network queue */
		nBufferSize = ulByteRcved;

		if( intface == 2 )
		{
		ZeroMemory(&pParent->m_addrSender, sizeof(SOCKADDR_IN) );

		nRet = recvfrom(pParent->m_sDiscovery, 
						pRcvBuffer, 
						nBufferSize, 
						0, 
						(SOCKADDR *)&pParent->m_addrSender, 
						&nAddrSize);
	
		if( SOCKET_ERROR  == nRet )
		{
			if( WSAECONNRESET == WSAGetLastError() )
			{
				/* do something properly */
				cout<< "[AMTCtrl] No connection" <<endl;
				continue;
			}
			else
			{
				/* we may need to retry based on error code */
				//cout << "receive error (" << WSAGetLastError() << ")" << endl;
				printf("receive error (%d)\n", WSAGetLastError()  );
			}
		}
		}
		else
		{
			nRet = recv(pParent->m_sUDP, pRcvBuffer, nBufferSize, 0 );
			if( SOCKET_ERROR  == nRet )
			{
				if( WSAECONNRESET == WSAGetLastError() )
				{
					/* do something properly */
					cout<< "[AMTCtrl] No connection" <<endl;
					continue;
				}
				else
				{
					/* we may need to retry based on error code */
					//cout << "receive error (" << WSAGetLastError() << ")" << endl;
					printf("receive error (%d)\n", WSAGetLastError()  );
				}
			}

		}
		
		

		/* put read data into our queue for processing */
		pParent->m_listMsgQ.push_back(pRcvBuffer);

		/* set signal */
		bRet = ::SetEvent(pParent->m_hRcvEvent);

		if( FALSE == bRet )
		{
			/* event signaling error */
			cout<< "SetEvent error ("<< GetLastError() << ")" << endl;
		}
		
	}// end of while loop ////////////////////////////////
}
#endif

void CAMTCtrl::MsgRcvProc(void* pParam)
{
	CAMTCtrl*        pParent; 
	int              nRet;
	unsigned long    ulByteRcved;
	char*            pRcvBuffer;
	int              nBufferSize;
	fd_set           readFDs;
	BOOL             bRet;
	RELAY_MSG        RelayMsg; 
	//HANDLE           hRcvEvent;

	/* initialization */
	pParent     = (CAMTCtrl*)pParam;
	ulByteRcved = 0;
	nBufferSize = 0;
	pRcvBuffer  = NULL;

	
	/* initialize fd set */
	FD_ZERO(&readFDs);
	FD_SET( pParent->m_sUDP, &readFDs );
	
	while(1)
	{
		/* wait here until get some messages */
		nRet = select( FD_SETSIZE, &readFDs, NULL, NULL, NULL );
		
		if( SOCKET_ERROR == nRet  )
		{
			DbgPrint("    [AMTCtrl] UDP Select error (%d)", WSAGetLastError() );
		}

		/* check size of receiving buffer */
		nRet = ioctlsocket( pParent->m_sUDP, FIONREAD, &ulByteRcved );

		/* allocate memory as much as received size and receive it */
		//pRcvBuffer  = new char[ulByteRcved];
		RelayMsg.pcData = NULL;
		RelayMsg.pcData = new char[ulByteRcved];

		if( NULL == RelayMsg.pcData )
		{

			/* system may out of memory - what shall we do?? */
			DbgPrint("    [AMTCtrl] UDP msg memory allocation failed");
			continue;
		}
		else
		{
#ifdef MEM_DEBUG
			char szMessage[1024];
			ZeroMemory( szMessage, sizeof(szMessage) );
			int nSize = sprintf_s(szMessage, "0x%08x@File: %s, Line:%d", RelayMsg.pcData, __FILE__, __LINE__);
			Mem_Debug(szMessage, nSize, TRUE );
#endif
		}

		RelayMsg.nDataLength = ulByteRcved;

		/* receive the data from network queue */
		nRet = recv(pParent->m_sUDP, RelayMsg.pcData, RelayMsg.nDataLength, 0 );

		if( SOCKET_ERROR  == nRet )
		{
			if( WSAECONNRESET == WSAGetLastError() )
			{
				/* do something properly */
				
				DbgPrint("    [AMTCtrl] UDP socket no connection");
#ifdef MEM_DEBUG
				char szMessage[1024];
				ZeroMemory( szMessage, sizeof(szMessage) );
				int nSize = sprintf_s(szMessage, "0x%08x@File: %s, Line:%d", RelayMsg.pcData, __FILE__, __LINE__);
				Mem_Debug(szMessage, nSize, FALSE );
#endif			
				delete [] RelayMsg.pcData;
					
				continue;
			}
			else
			{
				/* we may need to retry based on error code */
				DbgPrint("    [AMTCtrl] UDP msg receive error");
			}
		}

		//EnterCriticalSection( &g_csMsgQ );
		/* put read data into our queue for processing */
		// fast Mcast data prevent controlling message processing -> double q, MSG Q has higher processing priority
		///printf("\n[+] RELAY_MSG: 0x%x", RelayMsg.pcData );

		EnterCriticalSection(&g_csMsgQ);
		if( RelayMsg.pcData[0] == AMT_MCAST_DATA )
		{
			pParent->m_listForwardQ.push_back(RelayMsg);
		}
		else
		{
			pParent->m_listMsgQ.push_back(RelayMsg);
		}
		LeaveCriticalSection( &g_csMsgQ );

		/* set signal */
		bRet = ::SetEvent(pParent->m_hRcvEvent);
		//printf("\n [ + ]");
		if( bRet == 0 )
		{
			DbgPrint("    [AMTCtrl]Fail set event (%d)", GetLastError() );
		}
	}

}
 

/* ==============================================================
*
* Name: SendRelayDiscovery
* Parameters: void
* Return    : void
* Descriptions:  send relay disocvery message
*
===============================================================*/

void CAMTCtrl::SendRelayDiscovery( void )
{
	char          chaSendBuffer[8];
	unsigned int  ulNonce;
	int           nRet;

	/* initialization */
	ulNonce = 0;
	nRet    = 0;
	ZeroMemory( chaSendBuffer, sizeof(chaSendBuffer) );

	/* 
	 * create AMT disocvery message format 
	 * +---------------------------------------------------+
	 * | Msg Type(1Byte)| Reserved (3 byte)| nonce (4byte) |
	 * +---------------------------------------------------+ 
	 */

	chaSendBuffer[0] = AMT_RELAY_DISCOVERY;
	chaSendBuffer[1] = 0;
	chaSendBuffer[2] = 0;
	chaSendBuffer[3] = 0;


	/* create nounce */
	m_ulRelayNonce = rand();
	//printf("\n dico nonce: %d", m_ulRelayNonce);

	/* convert to network byte order */
	ulNonce  = htonl( m_ulRelayNonce );

	/* copy nonce into send buffer */
	memcpy( (void*)&chaSendBuffer[4], (void*)&ulNonce, sizeof(unsigned int) );

	/* send data */
	nRet = sendto( m_sDiscovery, //pParent->m_sDiscovery, 
		           chaSendBuffer, 
				   sizeof(chaSendBuffer), 
				   0, 
				   (SOCKADDR *)&m_addrDiscovery, //(SOCKADDR *)&pParent->m_addrDiscovery, 
				   sizeof(SOCKADDR) 
				   );		
	
	if( SOCKET_ERROR == nRet )
	{
		//printf("\n    [AMTCtrl]error at sending relay discovery (err: %d)", WSAGetLastError() );
		DbgPrint("    [AMTCtrl]error at sending relay discovery (err: %d)", WSAGetLastError() );
	}
	else
	{
		//printf("\n[AMTCtrl] Discovery msg has been sent");
		DbgPrint("[AMTCtrl] Discovery msg has been sent");
	}
}

/* ==============================================================
*
* Name: SendRelayRequest
* Parameters: 
* Return    : void
* Descriptions:  send a request
*
===============================================================*/
void CAMTCtrl::SendRelayRequest( unsigned int ulCurrentNonce )
{

	char          chaSendBuffer[8];
	unsigned int  ulNonce;
	int           nRet;
	int           nRetry;

	/* initialization */
	ZeroMemory( chaSendBuffer, sizeof(chaSendBuffer) );

	ulNonce = 0;
	nRet    = 0;
	nRetry  = 0;

	/* 
	 * create AMT request message format 
	 * +---------------------------------------------------+
	 * | Msg Type(1Byte)| Reserved (3 byte)| nonce (4byte) |
	 * +---------------------------------------------------+ 
	 */
	
	chaSendBuffer[0] = AMT_REQUEST;
	chaSendBuffer[1] = 0;
	chaSendBuffer[2] = 0;
	chaSendBuffer[3] = 0;

	/* copy nonce to request item */
	// do more work
	//SetCurrentNonce( nNonce );

	/* convert to network byte order */
	ulNonce  = htonl( ulCurrentNonce );

	/* copy nonce into send buffer */
	memcpy( (void*)&chaSendBuffer[4], (void*)&ulNonce, sizeof(unsigned int) );

	/* send data */

	while( nRetry < UDP_SEND_RETRY )
	{
#if 0
		SOCKADDR mySocketAddr;
		SOCKADDR_IN mySockAddrIn;
		int      nMySockAddrSize;
		
		nMySockAddrSize = sizeof(SOCKADDR);
		getsockname( m_sUDP, &mySocketAddr, &nMySockAddrSize);
		memcpy( &mySockAddrIn, &mySocketAddr, sizeof(SOCKADDR_IN) );

		DbgPrint("[AMTCtrl] My Port : %d" ,  ntohs( mySockAddrIn.sin_port ) );
#endif

		nRet = send( m_sUDP, chaSendBuffer, sizeof(chaSendBuffer), 0);
		
		if( SOCKET_ERROR == nRet )
		{
			nRetry++;
			//printf("\n    [AMTCtrl] Fail to send request (err: %d), retry: %d", WSAGetLastError(), nRetry );
			DbgPrint("    [AMTCtrl] Fail to send request (err: %d), retry: %d", WSAGetLastError(), nRetry );
			continue;
		}
		//printf("\n[AMTCtrl] GW->Relay (REQUEST)");
		//printf("\n[AMTCtrl] Send request (nonce: %d)", ulCurrentNonce );
		DbgPrint("[AMTCtrl] GW->Relay (REQUEST)");
		DbgPrint("[AMTCtrl] Send request (nonce: %d, 0x%x)", ulCurrentNonce, ulCurrentNonce );

		break;
	}
}

/* ==============================================================
*
* Name: SendMembership
* Parameters: 
* Return    : void
* Descriptions:  send a membership
*
===============================================================*/

void CAMTCtrl::SendMembership( unsigned char* pMAC,        unsigned long ulCurrentNonce, 
							   unsigned char* pIGMPPacket, int nPktLength )
{
	char*         pSendBuffer;
	int           nBufferSize;
	int           nRet;
	unsigned long ulNonce;
	int           nRetry;
	
	/* initialization */
	pSendBuffer = NULL;
	nBufferSize = 0;
	nRet        = 0;
	nRetry      = 0;

	/* allocate memory */
	nBufferSize = RESPONSE_MEMBERSHIP_HADER_LEN + nPktLength;
	pSendBuffer = new char[nBufferSize];
	
	if( NULL == pSendBuffer )
	{
		//cout<<"SendMembership() memory allocation fail"<<endl;
		//printf("\n    [AMTCtrl] SendMembership() memory allocation fail");
		DbgPrint("    [AMTCtrl] SendMembership() memory allocation fail");
		return;
		
	}
	else
	{
#ifdef MEM_DEBUG
		char szMessage[1024];
		ZeroMemory( szMessage, sizeof(szMessage) );
		int nSize = sprintf_s(szMessage, "0x%08x@File: %s, Line:%d", pSendBuffer, __FILE__, __LINE__);
		Mem_Debug(szMessage, nSize, TRUE );
#endif
	}

	ZeroMemory( pSendBuffer, nBufferSize );

	/* 
	 * create AMT request message format 
	 * +----------------------------------------------------------------------------------+
	 * | Msg Type(1 byte)| Reserved (1 byte)| MAC (6 byte)| nonce (4 byte) | IGMP packet  |
	 * +----------------------------------------------------------------------------------+
	 */
	
	pSendBuffer[0] = AMT_MEMBERSHIP_CHANGE;
	pSendBuffer[1] = 0; 

	/* copy relay MAC */
	memcpy( &pSendBuffer[2], pMAC, RESPONSE_MAC_LEN );

	/* copy nonce */
	ulNonce = htonl(ulCurrentNonce);
	memcpy( &pSendBuffer[8], &ulNonce, RESPONSE_NONCE_LEN );

	/* copy original IGMP packet */
	memcpy( &pSendBuffer[12], pIGMPPacket, nPktLength );

	/* send data */
	while( nRetry < UDP_SEND_RETRY )
	{
#if 0
		SOCKADDR mySocketAddr;
		SOCKADDR_IN mySockAddrIn;
		int      nMySockAddrSize;
		
		nMySockAddrSize = sizeof(SOCKADDR);
		getsockname( m_sUDP, &mySocketAddr, &nMySockAddrSize);
		memcpy( &mySockAddrIn, &mySocketAddr, sizeof(SOCKADDR_IN) );

		DbgPrint("[AMTCtrl] My Port : %d" ,  ntohs( mySockAddrIn.sin_port ) );
#endif
		
		nRet = send( m_sUDP, pSendBuffer, nBufferSize, 0);
		
		if( SOCKET_ERROR == nRet )
		{
			nRetry++;
			//printf("\n    [AMTctrl] Send memebership fail (err: %d) -> retry: %d", WSAGetLastError(), nRetry );
			DbgPrint("    [AMTctrl] Send memebership fail (err: %d) -> retry: %d", WSAGetLastError(), nRetry );
			continue;
		}
		//printf("\n[AMTCtrl] GW->Relay (MEMBERSHIP)");
		//printf("\n[AMTCtrl] Send Membership (nonce: %d)", ulCurrentNonce);
		DbgPrint("[AMTCtrl] GW->Relay (MEMBERSHIP)");
		DbgPrint("[AMTCtrl] Send Membership (nonce: %d, 0x%x)", ulCurrentNonce, ulCurrentNonce);

		break;
	}
	
	/* deallocate the memory */
	if( NULL != pSendBuffer )
	{
#ifdef MEM_DEBUG
		char szMessage[1024];
		ZeroMemory( szMessage, sizeof(szMessage) );
		int nSize = sprintf_s(szMessage, "0x%08x@File: %s, Line:%d", pSendBuffer, __FILE__, __LINE__);
		Mem_Debug(szMessage, nSize, FALSE );
#endif
		delete [] pSendBuffer;
	}
}

/* ==============================================================
*
* Name: ForwardingMcastData
* Parameters: 
* Return    : void
* Descriptions:  pass multicast data to the kernel
*
===============================================================*/
void CAMTCtrl::ForwardingMcastData( char* pPktBuffer, int nLength )
{
	BOOL bRet;
	int  nSkip;
	
	/* skip amt header (type 1 byte + reserve 1 byte) */
	nSkip = 2;

	/* forward it to kernel */
	bRet = m_pObjPktSender->SendPkt((unsigned char*)(pPktBuffer+nSkip), (unsigned int)(nLength-nSkip) );
	
#if 0
	if( FALSE == bRet )
	{
		//printf("\n[AMTCTRl] Fail to send packet");
		//m_pObjPktSender->SendPkt((unsigned char*)(pPktBuffer+nSkip), (unsigned int)(nLength-nSkip) );
	}
#endif
}


/* ==============================================================
*
* Name: AddRequest
* Parameters:
* Return:
* Descriptions:
* 1. add new request into the request queue
* 2. If return value is FALSE, caller should deallocate IGMP
*    buffer properly because it is not necessary
*
===============================================================*/
BOOL CAMTCtrl::AddRequest( REQUEST& newReq )
{
	
	//EnterCriticalSection( &g_csReqQ );

	m_listReqestQ.push_back(newReq);
	DbgPrint("[AMTCtrl] Add req: size = %d", m_listReqestQ.size());

	//LeaveCriticalSection( &g_csReqQ );
		
	/* process or not? - check current AMT status */
	/*
	if( RELAY_FOUND == m_nRelayStatus )
	{
		ProcessRequest();
	}
	*/
	if( RELAY_FOUND != m_nRelayStatus )
	{
		CreateDiscoveryTimer();
	}
	ProcessRequest();
	
	return TRUE;
}


/* ==============================================================
*
* Name: ProcessRequest
* Parameters:
* Return:
* Descriptions:
* 1. create new timer for a request so that it can be serviced
* 2. if request exceeds maximum number of request, 
*    remove it from queue and initiate relay discovery.
*
===============================================================*/

void CAMTCtrl::ProcessRequest()
{
	list<REQUEST>::iterator iter;
	BOOL bSuccess;
	REQ_INFO*    pReqInfo;

	/* enable request timer */
	// first, check timer handle. if it is running do nothing.
	// second, if handle is null, create one and keep it running

	EnterCriticalSection( &g_csReqQ );
	//printf("\n PR: enter cs");
	for( iter = m_listReqestQ.begin(); iter != m_listReqestQ.end(); iter++ )
	{
		
		if( NULL == iter->hReqTimerHndl )
		{
			pReqInfo = new REQ_INFO;
			///printf("\n[+]REQ_INFO: 0x%x", pReqInfo );
#ifdef MEM_DEBUG
			char szMessage[1024];
			ZeroMemory( szMessage, sizeof(szMessage) );
			int nSize = sprintf_s(szMessage, "0x%08x@File: %s, Line:%d", pReqInfo, __FILE__, __LINE__);
			Mem_Debug(szMessage, nSize, TRUE );
#endif
			
			/* timer callback parameter */
			pReqInfo->pParent = (void*)this; 
			pReqInfo->ulNonce = rand();

			/* set initial count */
			iter->ulReqCount = 0;

			/* assign nonce */
			iter->ulNonce = pReqInfo->ulNonce;

			iter->pReqInfo = pReqInfo;

			/* create timer and set its handle */
			bSuccess = CreateTimerQueueTimer( &iter->hReqTimerHndl,    // handle to timer
									          m_hTimerQueue,           // handle to timer queue
											  (WAITORTIMERCALLBACK)RequestTimerHandler, // timer callback function
											  (PVOID)pReqInfo,         // callback parameter
											  500,                       // timer due time
											  REQ_TIMER_INIT_PERIOD,   // timer period
											  //WT_EXECUTEINTIMERTHREAD  // options
											  WT_EXECUTELONGFUNCTION
											  );
			if( bSuccess == 0 )
			{
				printf("\n Failt set initial request timer: %d", GetLastError() );
			}
			
		}
#if 0
		else
		{
			
			/* check number of maximum trial */
			if( iter->ulReqCount >= GW_REQUEST_MAX )
			{
				/* delete timer */
				bSuccess = DeleteTimerQueueTimer( m_hTimerQueue, iter->hReqTimerHndl, INVALID_HANDLE_VALUE );
				if( FALSE == bSuccess )
				{
					//printf("\n    [AMTCtrl] Fail to delete req. timer (err: %d)", GetLastError() );
					DbgPrint("    [AMTCtrl] Fail to delete req. timer (err: %d)", GetLastError() );
				}

				if( NULL != iter->pReqInfo )
				{
					///printf("\n[-]REQ_INFO: 0x%x", iter->pReqInfo );
					delete iter->pReqInfo;
				}
				
				/* deallocate igmp request memory */
				if( NULL != iter->pucIGMPData )
				{
					///printf("\n[-]IGMP_REQ: 0x%x",iter->pucIGMPData);
					delete [] iter->pucIGMPData;
				}

				/* remove a request from queue */
				iter = m_listReqestQ.erase(iter);

				DbgPrint("[AMTCtrl] request processing reached to its MAX retry count - relay discovery start");

				/* start relay discovery */
				SetRelayStatus(RELAY_NOT_FOUND);
				CreateDiscoveryTimer();
			}

		}
#endif
	} // end of for loop /////////////////
	//printf("\n PR: leave cs");
	LeaveCriticalSection( &g_csReqQ );
}


/* ==============================================================
*
* Name: RequestTimerHandler
* Parameters:
* Return:
* Descriptions: Request timer handler proc.
*
===============================================================*/
void CAMTCtrl::RequestTimerHandler( PVOID lpParameter, BOOLEAN TimerOrWaitFired )
{
	CAMTCtrl*                  pParent;
	REQ_INFO*                  pReqInfo;
	//REQ_INFO*                  pReScheduledReqInfo;

	list<REQUEST>::iterator    reqQIter;
	unsigned int               ulNonce;
	
	BOOL                       bSuccess;
	int                        nDue;
	//int                        nPeriod;
	int                        nTempReqCount;
	HANDLE                     hTempTimerHandle;
	BOOL                       bFoundMatch;

	nTempReqCount = 0;
	bFoundMatch   = FALSE;
	bSuccess      = FALSE;


	pReqInfo = (REQ_INFO*)lpParameter;
	
	pParent  = (CAMTCtrl*)(pReqInfo->pParent);
	ulNonce  = pReqInfo->ulNonce;
	

    EnterCriticalSection( &g_csReqQ );
	//printf("\n RH: enter cs");
	for( reqQIter = pParent->m_listReqestQ.begin(); reqQIter != pParent->m_listReqestQ.end(); reqQIter++)
	{
		if( reqQIter->ulNonce == ulNonce )
		{
			/* increase request counter */
			reqQIter->ulReqCount++;

			/* copy required info to local for fast CS leaving */
			nTempReqCount    = reqQIter->ulReqCount;
			hTempTimerHandle = reqQIter->hReqTimerHndl;
			bFoundMatch      = TRUE;
			break;
		}
	}
	
	
	/* check maximum retry */
	//if( bFoundMatch == TRUE && nTempReqCount > GW_REQUEST_MAX )
	if( bFoundMatch == TRUE )
	{
		if( nTempReqCount > GW_REQUEST_MAX+1 )
		{
			DbgPrint("[AMTCtrl] Request expired - remove (nonce: %d)", ulNonce);

			/* delete timer */
			bSuccess = DeleteTimerQueueTimer( pParent->m_hTimerQueue, hTempTimerHandle, NULL );
			if( FALSE == bSuccess )
			{
				int nError = GetLastError();
				if( ERROR_IO_PENDING != nError )
				{
					DbgPrint("    [AMTCtrl] Fail to delete req. timer (err: %d)", nError );
				}
			}

			if( NULL != reqQIter->pReqInfo )
			{
				///printf("\n[-]REQ_INFO: 0x%x", reqQIter->pReqInfo );
	#ifdef MEM_DEBUG
				char szMessage[1024];
				ZeroMemory( szMessage, sizeof(szMessage) );
				int nSize = sprintf_s(szMessage, "0x%08x@File: %s, Line:%d", reqQIter->pReqInfo, __FILE__, __LINE__);
				Mem_Debug(szMessage, nSize, FALSE );
	#endif
				delete reqQIter->pReqInfo;
			}
			
			/* deallocate igmp request memory */
			if( NULL != reqQIter->pucIGMPData )
			{
				///printf("\n[-]IGMP_REQ: 0x%x",reqQIter->pucIGMPData);
	#ifdef MEM_DEBUG
				char szMessage[1024];
				ZeroMemory( szMessage, sizeof(szMessage) );
				int nSize = sprintf_s(szMessage, "0x%08x@File: %s, Line:%d", reqQIter->pucIGMPData, __FILE__, __LINE__);
				Mem_Debug(szMessage, nSize, FALSE );
	#endif
				delete [] reqQIter->pucIGMPData;
			}

			/* remove a request from queue */
			reqQIter = pParent->m_listReqestQ.erase(reqQIter);

			bFoundMatch = FALSE;

		}
		else if( nTempReqCount > GW_REQUEST_MAX )
		{
			DbgPrint("[AMTCtrl] MAX retry count - relay discovery start");

			pParent->SetRelayStatus(RELAY_NOT_FOUND);
			pParent->CreateDiscoveryTimer();

			//bFoundMatch = FALSE;	
		}
	}
	//printf("\n RH: leave cs");
	LeaveCriticalSection( &g_csReqQ );	
	
	if( TRUE == bFoundMatch )
	{
		if(nTempReqCount > GW_REQUEST_MAX)
		{
			DbgPrint("[AMTCtrl] keep an expired request for late query arrival");
			nDue = GW_REQUEST_RETRY_INTERVAL * 3 + 2000; // 5 sec

		}
		else
		{
			/* send request */
			DbgPrint("[AMTCtrl] req. cnt = %d", nTempReqCount);
			pParent->SendRelayRequest( ulNonce );

			/* adjust timer period - binanry exponential back off */
			//nDue    = ( 1 << (nTempReqCount) ) * 1000;
			//nDue = ((REQUEST_SEND_INTERVAL + nTempReqCount ) * 1000);
			nDue = GW_REQUEST_RETRY_INTERVAL;
		}

		bSuccess = ChangeTimerQueueTimer( pParent->m_hTimerQueue, 
										  hTempTimerHandle, //reqQIter->hReqTimerHndl, 
										  nDue,
										  REQ_TIMER_INIT_PERIOD 
										  );
		
		DbgPrint("[AMTCtrl] next timer: (+)%d ms", nDue);
		if( FALSE == bSuccess )
		{
			DbgPrint("    [AMTCtrl] fail to change request timer" );
		}

	} // found match IF statement
	//else
	//{
		
		//DbgPrint("[AMTCtrl] non match req. or retry max req.");
	//}

	

}