/*
* AMT Gateway Win32 version
*
* THE UNIVERSITY OF TEXAS AT DALLAS, ANDES Lab.
*
* Author: Juhyun Oh (Justin), justinoh@student.utdallas.edu
*
*/

#pragma once

#include "common.h"
#include "PktSender.h"

#if 0
#include "CurrentGroupInfo.h"
#endif

class CAMTCtrl
{
public:
	/* constructor, destructor */
	CAMTCtrl(void);
	~CAMTCtrl(void);

	/* user defined pubic methods */
	static void DiscoveryRcvProc( void* );	 // discovery message receiving thread 
	static void MsgRcvProc( void* );         // AMT message receiving thread
	static void AMTProtocolHandler( void* ); // AMT message parsing and handling thread
	
	static void CALLBACK DiscoveyTimerHandler( PVOID, BOOLEAN );   // dicovery timer handler
	static void CALLBACK RequestTimerHandler( PVOID, BOOLEAN );    // request timer handler
	static void CALLBACK IGMPQueryTimerHandler( PVOID, BOOLEAN );  // IGMP query timer handler
	
	void SetPktSenderObject( CPktSender* ); 
#if 0
	void SetGroupInfoObject( CCurrentGroupInfo* ); 
#endif
	//void SetIFAddress( string ); // get current default interface address IPv4 string
	BOOL AddRequest( REQUEST& ); // request add interface

	static CRITICAL_SECTION g_csReqQ; // crtical section for request queue
	static CRITICAL_SECTION g_csMsgQ; // critical section for AMT message queue


	//void CreateIGMPQueryTimer( void );
protected:
	
	/* initialization and clean up */
	void InitializeAMTProtocolHandler( void );
	void InitializeSocket( void );
	void InitializeReceiver( void );
	void InitializeTimerQueue( void );
	
	// discovery timer
	void CreateDiscoveryTimer( void );
	void DeleteDiscoveryTimer( void );

	// periodic IGMP querier
	void CreateIGMPQueryTimer( void );
	void DeleteIGMPQueryTimer( void );
	
	// relay status
	void SetRelayStatus( int );
	int  GetRelayStstus( void );

	// AMT message processing 
	void SendRelayDiscovery( void );
	void SendRelayRequest( unsigned int );
	void SendMembership( unsigned char*, unsigned long, unsigned char*, int );
	void ForwardingMcastData( char*, int );
	void OnAMTAdvertisementMsg( char* );
	void OnAMTMembershipQueryMsg( char* );

	// request handling 
	void ProcessRequest();

	// configuration file read 
	BOOL ReadConfigFile(void); 

	//unsigned short GetChecksum(unsigned short *pusIPBuffer, int nLen);

	/* */
	//int GetRandom(void);

private:

	SOCKET				m_sUDP;          // UDP socket
	SOCKET				m_sDiscovery;    // relay discovery socket
	SOCKET				m_sQuery;
	SOCKADDR_IN			m_addrDiscovery;
	SOCKADDR_IN			m_addrSender;    // sender address
	SOCKADDR_IN			m_addrRelay;     // relay unicast address
	SOCKADDR_IN			m_stLocalAddr;   // local address
	SOCKADDR_IN			m_stSvrAddr;	 // server address

	HANDLE				m_hRcvThread;          // UDP receiver thread handle
	HANDLE				m_hDiscoveryRcvThread; //discovery receiver thread handle
	HANDLE				m_hAMTHandlerThread;   // receiver thread handle
	HANDLE				m_hRcvEvent;           // data reception event handle
	
	
	HANDLE				m_hTimerQueue;         // timerQ handle
	HANDLE				m_hDisoveryTimer;      // discovery timer handle
	HANDLE				m_hIGMPQueryTimer;     // IGMP query timer handle
	
	list<RELAY_MSG>		m_listMsgQ;     // the storage which holds received data
	list<RELAY_MSG>		m_listForwardQ; // the storage which holds received data
	list<REQUEST>		m_listReqestQ;  // request queue

	unsigned long		m_ulRelayNonce; // relay discovery nonce
	int					m_nRelayStatus; // current relay status

	CPktSender*			m_pObjPktSender; //packet sender object
#if 0
	CCurrentGroupInfo*	m_pObjGroupInfo; // group info object
#endif

	string				m_strRelayIP;
	string				m_strNicIP;

};

