#include "CurrentGroupInfo.h"

CRITICAL_SECTION CCurrentGroupInfo::g_csGroupInfo;

CCurrentGroupInfo::CCurrentGroupInfo(void)
{
	InitializeCriticalSection(&g_csGroupInfo);
	
	// table size ( number of rows)
	m_nSize = 0;

	// initialy set IGMP version to 0 neither version 2 or 3
	m_unInitialIGMPVer = 0;

	// initilize packet vector index
	m_nPktVecIndex = 0;

}

CCurrentGroupInfo::~CCurrentGroupInfo(void)
{
	DeleteCriticalSection(&g_csGroupInfo);
}

int CCurrentGroupInfo::GetSize()
{
	return m_nSize;
}

BOOL CCurrentGroupInfo::AddGroup( unsigned long ulGroup, unsigned long ulSource )
{
	//BOOL bNewGroup = TRUE;
	// insert to group entry table
	
	//for logging.......
	IN_ADDR	stAddrGroup, stAddrSource;
	stAddrGroup.S_un.S_addr  = ulGroup;
	stAddrSource.S_un.S_addr = 	ulSource;	
		
	string strSrc = inet_ntoa(stAddrSource);
	string strGrp = inet_ntoa(stAddrGroup);


	list<vector<unsigned long>>::iterator	tblIter;
	BOOL bGroupMatch = FALSE;
	BOOL bSrcMatch  = FALSE;

	// 1. search table and find matching group
	for( tblIter = m_listGroupEntryTbl.begin(); tblIter != m_listGroupEntryTbl.end(); tblIter++ )
	{
		// compare with group address
		// iterator is a pointer of a row
		// group is always at the 0 index of row
		bSrcMatch  = FALSE;

		if( ulGroup == tblIter->at(0) )
		{
			bGroupMatch = TRUE;

			// 2. found the group in the table, then search the vector to see if there is duplicate item
			int nSize = tblIter->size();
			
			for( int i = 1; i < nSize; i++ )
			{
				if( tblIter->at(i) == ulSource )
				{
					// found the same source address
					bSrcMatch = TRUE;
					break;
				}
			}

			if( FALSE == bSrcMatch )
			{
				// add source to current group row
				tblIter->push_back(ulSource);
				break; // out of main loop
			}
			else
			{
				DbgPrint("[GroupInfo] Add (Duplicated): (S,G) = (%s, %s)", strSrc.c_str(), strGrp.c_str() );
			}
		}
		
	}

	if( FALSE == bGroupMatch )
	{
		// no matching entry in the table
		// create a row of the table
		// +------------------------------------------------------
		// | Group | Source 1 |
		// +------------------------------------------------------
		vector<unsigned long> vectTblRow;
		vectTblRow.push_back( ulGroup );
		vectTblRow.push_back( ulSource );

		// add the row to table
		m_listGroupEntryTbl.push_back(vectTblRow);

		DbgPrint("[GroupInfo] Add: (S,G) = (%s, %s)", strSrc.c_str(), strGrp.c_str() );
	}

	m_nSize = m_listGroupEntryTbl.size();


	if( bGroupMatch == TRUE &&  bSrcMatch == TRUE )
		return FALSE;
	
	return TRUE;

}

BOOL CCurrentGroupInfo::DeleteGroup(unsigned long ulGroup, unsigned long ulSource)
{

	//for logging.......
	IN_ADDR	stAddrGroup, stAddrSource;
	stAddrGroup.S_un.S_addr  = ulGroup;
	stAddrSource.S_un.S_addr = 	ulSource;	
		
	string strSrc = inet_ntoa(stAddrSource);
	string strGrp = inet_ntoa(stAddrGroup);


	list<vector<unsigned long>>::iterator	tblIter;
	//BOOL bGroupMatch = FALSE;
	
	tblIter = m_listGroupEntryTbl.begin();

	BOOL bSrcMatch = FALSE;

	// 1. search table and find matching group
	while( tblIter != m_listGroupEntryTbl.end() )
	{
		// compare with group address
		// iterator is a pointer of a row
		// group is always at the 0 index of row
		bSrcMatch  = FALSE;

		if( ulGroup == tblIter->at(0) )
		{
			// 2. found the group in the table, then search the vector to see if there is duplicate item
			vector<unsigned long>::iterator  rowIter;
			
			for( rowIter = (tblIter->begin()+1); rowIter != tblIter->end(); rowIter++ )
			{
				if( *rowIter == ulSource )
				{
					// found the same source address
					bSrcMatch = TRUE;
					break;
				}
			}

			if( TRUE == bSrcMatch )
			{
				// erase source
				tblIter->erase(rowIter);

				// no source entry in a row, group address should also be removed
				if( tblIter->size() < 2 )
				{
					tblIter = m_listGroupEntryTbl.erase(tblIter);
				}

				DbgPrint("[GroupInfo] Delete : (S,G) = (%s, %s)", strSrc.c_str(), strGrp.c_str() );

				// out of loop
				break;

			}
			
		}// end of table iter if

		tblIter++;
		
	}// end of while loop

	
	m_nSize = m_listGroupEntryTbl.size();

	return bSrcMatch;

}

void CCurrentGroupInfo::DeleteAllGroup(unsigned long ulGroup)
{

	// remove group in the table
	list<vector<unsigned long>>::iterator	tblIter;
	
	tblIter = m_listGroupEntryTbl.begin();

	while( tblIter != m_listGroupEntryTbl.end() )
	{
		if( tblIter->at(0) == ulGroup )
		{
			tblIter = m_listGroupEntryTbl.erase(tblIter);
			break;
		}

		tblIter++;
	}

	// not yet update size in here
	// m_nSize = m_listGroupInfo.size();
}

#if 0
void CCurrentGroupInfo::VerifyIGMPVersion( unsigned int unIGMPVer )
{
	if( 0x00 == m_unInitialIGMPVer )
	{
		m_unInitialIGMPVer  = unIGMPVer;
		return;
	}

	// IGMP v3 -> v2, once host start talking to IGMP v2 it will not go up to the IGMPv3
	if( IGMP_VER3 == m_unInitialIGMPVer && 
		IGMP_VER2 == unIGMPVer )
	{
		m_unInitialIGMPVer = IGMP_VER2;
		// delete the all (S,G) in the list and set (*,G)
		list<vector<unsigned long>>::iterator	tblIter;
		
		for( tblIter = m_listGroupEntryTbl.begin(); tblIter != m_listGroupEntryTbl.end(); tblIter++ )
		{
			// for every group remove all source and add 0.0.0.0
			while( tblIter->size() != 1 )
			{
				tblIter->pop_back();
			}

			// add 0.0.0.0
			tblIter->push_back(0);
		}

#if _DEBUG
	printf("\n IGMPv3 ----> IGMPv2");
	DumpTable();
#endif

	} // end of if


}
#endif

#if 0
void CCurrentGroupInfo::UpdateGroupInfo( unsigned char* puchOrgIGMPPkt, int nLen )
{
	unsigned char uchIGMPType;
	unsigned char* puchIGMPPkt;


	::EnterCriticalSection(&g_csGroupInfo);

	/* validate the pointer */
	if( NULL == puchOrgIGMPPkt || 0 >= nLen)
	{
		// do nothing
		DbgPrint("    [GroupInfo] invalid IGMP pkt pointer or length");
		return;
	}

	puchIGMPPkt = new unsigned char[nLen];
	if( NULL == puchIGMPPkt )
	{
		DbgPrint("    [GroupInfo] Cannot allocate memory");
		return;
	}
	// copy to temporary memory
	memcpy( puchIGMPPkt, puchOrgIGMPPkt, nLen );


	uchIGMPType = puchIGMPPkt[0];

	switch( uchIGMPType )
	{
		case IGMPV2_TYPE_JOIN:
			{
				VerifyIGMPVersion( IGMP_VER2 );

				unsigned long ulGroup;
				memcpy( &ulGroup, &puchIGMPPkt[4], sizeof(unsigned long) );
				AddGroup( ulGroup, 0 );
			}
			break;
		case IGMPV2_TYPE_LEAVE:
			{
				VerifyIGMPVersion( IGMP_VER2 );

				unsigned long ulGroup;
				memcpy( &ulGroup, &puchIGMPPkt[4], sizeof(unsigned long) );
				DeleteGroup( ulGroup, 0 );
			}
			break;
		case IGMPV3_TYPE_REPORT:
			{
				VerifyIGMPVersion( IGMP_VER3 );

				// get number of records
				unsigned short usNumRecords;
				memcpy( &usNumRecords, &puchIGMPPkt[6], sizeof(unsigned short) );

				// change to small edian
				unsigned int unRecords = (unsigned int)ntohs(usNumRecords);

				// iterate and include or exclude groups
				unsigned int unRecOffset = IGMPV3_GRPREC_OFFSET; // the first group record offset

				unsigned char	uchRecType    = 0;
				unsigned int	unAuxDataSize = 0;
				unsigned short	usNumSource   = 0;
				unsigned int	unNumSource   = 0;
				unsigned int	unTotRecSize  = 0;
				unsigned long	ulRecGroup    = 0;
				

				for( unsigned int i = 0; i < unRecords; i++ )
				{
					// get type - include, exclude
					//uchRecType = puchIGMPPkt[unRecOffset+IGMPV3_GRPREC_OFFSET];
					uchRecType = puchIGMPPkt[unRecOffset]; // very first of the each group record

					// get the length of current record
					// 1. get aux data length - AUX (1 byte)
					unAuxDataSize = (unsigned int)puchIGMPPkt[unRecOffset + IGMPV3_REC_AUXLEN_OFFSET];

					// 2. get number of source - Number of source (2 byte)
					memcpy( &usNumSource, &puchIGMPPkt[unRecOffset + IGMPV3_REC_NUMSRC_OFFSET], sizeof(unsigned short) );
					unNumSource = (unsigned int)ntohs(usNumSource);

					// 3. get toal size of one record

					// get group address
					memcpy( &ulRecGroup, &puchIGMPPkt[unRecOffset + IGMPV3_REC_GROUP_OFFSET], sizeof(unsigned long) );

					//              header 4 byte         + ( Num of src + 1 group ) * 4 byte
					unTotRecSize = IGMPV3_REC_HEADER_SIZE + (unNumSource + 1) * sizeof(unsigned long); 


					if( unNumSource > 0 )
					{
					
						// 4. iterates each source item
						unsigned int unRecSrcOffset = 0;
						for( unsigned int j = 0; j <  (unsigned int)unNumSource; j++ )
						{
							unsigned int ulRecSource = 0;
							memcpy( &ulRecSource, &puchIGMPPkt[unRecOffset + IGMPV3_REC_SOURCE_OFFSET + unRecSrcOffset], sizeof(unsigned long) );

							DbgPrint("[GroupInfo] Mode: %d", uchRecType);
							switch(uchRecType)
							{
								
								case IGMPV3_MODE_IS_INCLUDE: 
								case IGMPV3_ALLOW_NEW_SOURCES: 
									AddGroup( ulRecGroup, ulRecSource );
									break;
								case IGMPV3_MODE_IS_EXCLUDE:
								case IGMPV3_BLOCK_OLD_SOURCES:
									DeleteGroup( ulRecGroup, ulRecSource );
									break;
								case IGMPV3_CHANGE_TO_EXCLUDE_MODE:
								case IGMPV3_CHANGE_TO_INCLUDE_MODE:
									// delete all (*,G) and add (S,G)
									DeleteAllGroup(ulRecGroup);
									AddGroup( ulRecGroup, ulRecSource );
									break;
							}
							// go to next source						
							unRecSrcOffset = unRecSrcOffset + sizeof( unsigned long );
						} // end of for j
					}
					else
					{
						unsigned int ulRecSource = 0;
						DbgPrint("[GroupInfo] Mode: %d", uchRecType);
						switch(uchRecType)
						{
							case IGMPV3_MODE_IS_INCLUDE: 
							case IGMPV3_ALLOW_NEW_SOURCES: 
								AddGroup( ulRecGroup, ulRecSource );
								break;
							case IGMPV3_MODE_IS_EXCLUDE:
							case IGMPV3_BLOCK_OLD_SOURCES:
								DeleteGroup( ulRecGroup, ulRecSource );
								break;
							case IGMPV3_CHANGE_TO_EXCLUDE_MODE:
							case IGMPV3_CHANGE_TO_INCLUDE_MODE:
								// delete all (*,G) and add (S,G)
								DeleteAllGroup(ulRecGroup);
								AddGroup( ulRecGroup, ulRecSource );
								break;
						}
					} // end of number of source if

					// pointer of the next record
					unRecOffset = unRecOffset + unTotRecSize;
				}
			
			}
			break;
		default:
			DbgPrint("    [GroupInfo] invalid IGMP mmembership report");
			break;

	}// end of switch

	if( puchIGMPPkt != NULL )
	{
		delete [] puchIGMPPkt;
	}
#if _DEBUG
	DumpTable();
#endif

	::LeaveCriticalSection(&g_csGroupInfo);
	
}
#endif

void CCurrentGroupInfo::DumpTable()
{
	printf("\n============================================");
	printf("\n    Current group info table");
	printf("\n============================================");
	
	list<vector<unsigned long>>::iterator	tblIter;
	vector<unsigned long>::iterator			rowIter;

	for( tblIter = m_listGroupEntryTbl.begin(); tblIter != m_listGroupEntryTbl.end(); tblIter++ )
	{
		for( rowIter = (tblIter->begin()+1); rowIter != tblIter->end(); rowIter++ )
		{
			IN_ADDR	stAddrGroup, stAddrSource;
			stAddrGroup.S_un.S_addr  = tblIter->at(0);
			stAddrSource.S_un.S_addr = *rowIter;	
			
			string strSrc = inet_ntoa(stAddrSource);
			string strGrp = inet_ntoa(stAddrGroup);

			printf("\n (%s, %s)", strSrc.c_str(), strGrp.c_str() );
		}
	}


	printf("\n============================================");

}


#if 0
/*
* return: number of packet created
*/


int CCurrentGroupInfo::GetIGMPPacket(void)
{
	
	unsigned int				 unCurrentIGMPVer;
	list<vector<unsigned long>>  listGroupTblSS;
	int nNumOfPkt = -1;

	// work with the snapshot table - m_listGroupTblSS
	// 1. check igmp packet vector, and see if someting remain.
	if( FALSE == m_vecIGMPQuery.empty() )
	{
		vector<IGMP_QUERY_PKT>::iterator	igmpIter;
		
		for( igmpIter = m_vecIGMPQuery.begin(); igmpIter != m_vecIGMPQuery.end(); igmpIter++ )
		{
			// deallocate memory that are allocated in the previous call
			if( NULL != (unsigned char*)(igmpIter->puchBuffer) )
			{
				delete [] (unsigned char*)(igmpIter->puchBuffer);
			}
		}
		
		// clean up vector
		m_vecIGMPQuery.clear();
	}

	// reset IGMP packet vector index
	m_nPktVecIndex = 0;

	// 2. make snap shot of the table for synchronization
	
	::EnterCriticalSection(&g_csGroupInfo);
	
	listGroupTblSS   = m_listGroupEntryTbl;
	unCurrentIGMPVer = m_unInitialIGMPVer;

	::LeaveCriticalSection(&g_csGroupInfo);

	// get table size
	int nSnapShotTblSize = listGroupTblSS.size();

	// assemble IP header first
	// checksum index = 10, 11
	// Source ip index = 12,13,14,15
	// Destination ip: 224.0.0.1
	unsigned char achIPHeader[] = { 0x45, 0xc0, 0x00, 0x1c, 0xe1, 0x11, 0x00, 0x00, 
							        0x01, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
							        0xE0, 0x00, 0x00, 0x01 };
	

	// copy IP source address
	memcpy( &achIPHeader[12], &m_ulIFAddr, sizeof(unsigned long) );
	
	//compute checksum
	unsigned short usIPChecksum = GetChecksum( (unsigned short*)achIPHeader, sizeof(achIPHeader) );
	memcpy( &achIPHeader[10], &usIPChecksum, sizeof(unsigned short) );


	// 3. check current IGMP version
	if( IGMP_VER2 == unCurrentIGMPVer )
	{
		// genreate IGMPv2 query - general
		const unsigned char auchIGMPBuffer[] = {0x11, 0x64, 0xee, 0x9b, 0x00, 0x00, 0x00, 0x00};
		unsigned char* pIGMPPkt = new unsigned char[sizeof(achIPHeader) + sizeof(auchIGMPBuffer)];
		
		if( NULL != pIGMPPkt )
		{
			// copy IP header
			memcpy(pIGMPPkt, achIPHeader, sizeof(achIPHeader));
			// copy IGMP
			memcpy(pIGMPPkt+sizeof(achIPHeader), auchIGMPBuffer, sizeof(auchIGMPBuffer));

			IGMP_QUERY_PKT stIgmpQuery;
			stIgmpQuery.puchBuffer = pIGMPPkt;
			stIgmpQuery.nLen = sizeof(achIPHeader) + sizeof(auchIGMPBuffer);
			
			m_vecIGMPQuery.push_back(stIgmpQuery);
			nNumOfPkt = 1;
		}

	}
	else if( IGMP_VER3 == unCurrentIGMPVer )
	{
		// genreate IGMPv3 query for each group
		const unsigned char auchIGMPTemplate[] = { 0x11, 0x64, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
											       //0x02, 0x7d, 0x00, 0x00 };

		/* create packet based on the (S,G) table */
		list<vector<unsigned long>>::iterator ssTblIter;

		for( ssTblIter = listGroupTblSS.begin(); ssTblIter != listGroupTblSS.end(); ssTblIter++ )
		{
			int nNumSrc = ssTblIter->size() - 1;
			//int nMemSize =  sizeof(achIPHeader) + sizeof(auchIGMPTemplate) + nNumSrc * sizeof(unsigned long);
			int nMemSize =  sizeof(achIPHeader) + sizeof(auchIGMPTemplate);
			
			unsigned char* pIGMPv3Pkt = new unsigned char[nMemSize];
			
			if( NULL != pIGMPv3Pkt )
			{
				// copy IP
				memcpy( pIGMPv3Pkt, achIPHeader, sizeof(achIPHeader) );
				// copy IGMP
				memcpy( pIGMPv3Pkt+sizeof(achIPHeader), auchIGMPTemplate, sizeof(auchIGMPTemplate) );
				// set group address
				memcpy( pIGMPv3Pkt+sizeof(achIPHeader)+4, &(ssTblIter->at(0)), sizeof(unsigned long) );
				// copy source address
				/*
				for( int i = 1; i < nNumSrc; i++ )
				{
					// copy a source
					memcpy( pIGMPv3Pkt+sizeof(achIPHeader)+sizeof(auchIGMPTemplate)+((i-1)*sizeof(unsigned long)),
						    &(ssTblIter->at(i)),
							sizeof(unsigned long) );
				}
				*/

				// compute checksum
				unsigned short usChecksum = GetChecksum( (unsigned short*)(pIGMPv3Pkt + sizeof(achIPHeader)),
						                                 sizeof(auchIGMPTemplate) );
				// copy checksum
				memcpy( pIGMPv3Pkt+sizeof(achIPHeader)+2, &usChecksum, sizeof(unsigned short) );

				// put the packet into the container
				IGMP_QUERY_PKT stIgmpQuery;
				stIgmpQuery.puchBuffer = pIGMPv3Pkt;
				stIgmpQuery.nLen = nMemSize;
			
				m_vecIGMPQuery.push_back(stIgmpQuery);
				//m_vecIGMPQuery.push_back(pIGMPv3Pkt);
			}
			else
			{
				DbgPrint("    [GroupInfo] Fail to allocate memory");
				nNumOfPkt = -1;
				return nNumOfPkt;
			}

		}
		
		nNumOfPkt = listGroupTblSS.size();
	}
	else
	{
		DbgPrint("    [GroupInfo] Don't have record to generate IGMP packet");
		return 0;
	}


	return nNumOfPkt;


}

BOOL CCurrentGroupInfo::NextPacket( unsigned char** ppIGMPPkt, int* pLen )
{
	try
	{
		IGMP_QUERY_PKT stIGMP = m_vecIGMPQuery.at(m_nPktVecIndex);
		*ppIGMPPkt = stIGMP.puchBuffer;
		*pLen      = stIGMP.nLen;

		m_nPktVecIndex++;
	}
	catch( int nEx )
	{
		DbgPrint("    [GroupInfo] out of boundary (%d)", nEx );
		return FALSE;
	}
		
	return TRUE;
}
#endif

void CCurrentGroupInfo::SetIFAddress( string strIFAddr )
{
	//m_strCurrentIF = strIFAddr;
	m_ulIFAddr = inet_addr( strIFAddr.c_str() );
}

#if 0
unsigned short CCurrentGroupInfo::GetChecksum(unsigned short *pusIPBuffer, int nLen)
{
	int nleft = nLen;
    int sum = 0;
    unsigned short *w = pusIPBuffer;
    unsigned short answer = 0;

    while (nleft > 1) 
	{
        sum += *w++;
        nleft -= 2;
    }
    if (nleft == 1) 
	{
        *(unsigned char*)(&answer) = *(unsigned char*)w;
        sum += answer;
    }
    sum = (sum >> 16) + (sum & 0xffff);
    answer = ~sum;
    return (answer);
}
#endif

int CCurrentGroupInfo:: GetSourceSize( int nGrpIndex )
{

	list<vector<unsigned long>>::iterator	tblIter;

	tblIter = m_listGroupEntryTbl.begin();
	
	for( int i = 0; i < nGrpIndex; i++ )
		tblIter++;

	return (tblIter->size()-1);
}

ULONG CCurrentGroupInfo::GetSrc( int nGrpIndex, ULONG* pOutSrc )
{
	list<vector<unsigned long>>::iterator	tblIter;
	vector<unsigned long>::iterator			rowIter;
	ULONG ulGroupAddr;

	tblIter = m_listGroupEntryTbl.begin();
	
	for( int i = 0; i < nGrpIndex; i++ )
		tblIter++;

	int nOutIndex = 0;
	
	ulGroupAddr = *(tblIter->begin());
	
	for( rowIter = (tblIter->begin()+1); rowIter != tblIter->end(); rowIter++, nOutIndex++ )
	{
		pOutSrc[nOutIndex] = *rowIter;
	}

	return ulGroupAddr;

}
