/*
* AMT Gateway Win32 version
* 
* File: GatewayMain.cpp
*
* Description: main of the application. Initialize required subsystems and create objects
* to handle the AMT protocol
*
* THE UNIVERSITY OF TEXAS AT DALLAS, ANDES Lab.
* Juhyun Oh (Justin), justinoh@student.utdallas.edu
*
*/

#include "common.h"
#include "AMTCtrl.h"
#include "PktMonitor.h"
#include "PktSender.h"
#include <iphlpapi.h>
#include "CurrentGroupInfo.h"
#include "XMLParserCtrl.h"
#include "IGMPPktGen.h"

#include <Ws2tcpip.h>
#include <detours.h>


#include <conio.h>


using namespace std;

struct group_source_req
{
    unsigned char gsr_interface;
    struct sockaddr_storage gsr_group;
    struct sockaddr_storage gsr_source;
};

/////////////////////////////////////////////////////////////////////
// Global data

// shared memroy object name
//char szName[]="Global\\AMTSharedMemObj";
char g_szWMPath[PATH_BUF_SIZE];
char g_szVLCPath[PATH_BUF_SIZE];
char g_szQTPath[PATH_BUF_SIZE];
// detour
char g_szDetourDLLPath[1024];
char g_szDetourMarkerDLLPath[1024];


//HWND ghwndMemLogger;
//CRITICAL_SECTION CPktMonitor::g_csTempReqQ;


BOOL HaveAnotherInstance( void )
{
	BOOL bRet = FALSE;
	HANDLE hMutex = NULL;
	hMutex = CreateMutex(NULL,TRUE,TEXT("_WIN32_AMT_GATEWAY_MUTEX_"));
	
	if( GetLastError() == ERROR_ALREADY_EXISTS )
	{
		bRet = TRUE;
	}

	return bRet;
}
#if 0
	////////////////////////////////////////////////////////////
	// Create event for path set notification
	HANDLE hPathSet;
	hPathSet = ::CreateEvent( NULL,                       // security attribute (default)
		                      FALSE,                      // reset type (manual reset)
				              FALSE,                      // initial state (not signaled)
				              TEXT("_AMT_PATH_PARM_SET_") ); // event name
	if( NULL == hPathSet )
	{
		DbgPrint("    [GWMain] Fail to create SM event (err: %d)", GetLastError() );
		//_getch();
		return 0;
	}


	//HANDLE hShareMem = 0;
	//LPVOID pBuf      = NULL;
	BOOL   bSMWatch  = FALSE;

	///////////////////////////////////////////////////////////////
	// Check another instance
	if( HaveAnotherInstance() )
	{
		// 1. open shared memory
		// 2. write .amt file path
		if( WritePath2SharedMemory( argv[1], strlen(argv[1]) ) == TRUE )
		{
			// 3. set event and exit
			// do I have to check fail? what I can do in case of failure?
			::SetEvent( hPathSet );
			//wait few millisec
			::Sleep(100);
			::CloseHandle(hPathSet);
		}
		// exit silently
		return 0;
	}


	/////////////////////////////////////////////////////////////
	// Create event for request handling
	HANDLE  hReqAdd;
	hReqAdd = ::CreateEvent( NULL,                       // security attribute (default)
		                     FALSE,                      // reset type (manual reset)
				             FALSE,                      // initial state (not signaled)
				             TEXT("_REQ_ADD_SIGNAL_") ); // event name
				             
	if( NULL == hReqAdd )
	{
		DbgPrint("    [GWMain] Fail to create req add event (err: %d)", GetLastError() );
		return 0;
	}
#endif


#if 0
	/////////////////////////////////////////////////////////////
	// Launch shared memory watch thread
	// required info: REQUEST handle, 

	PATH_INFO stFilePathInfo;
	stFilePathInfo.hSMEvent = hPathSet;
	stFilePathInfo.pGroupInfoObj = (LPVOID)pObjGroupInfo;

	_beginthread( SMWatchProc, 0, (void*)&stFilePathInfo );
#endif

#if 0

PVOID CreateSharedMemory( HANDLE* phSharedMem )
{
	HANDLE	hMapFile;
	LPCTSTR pBuf;

	hMapFile = CreateFileMapping(
				 INVALID_HANDLE_VALUE,    // use paging file
				 NULL,                    // default security 
				 PAGE_READWRITE,          // read/write access
				 0,                       // max. object size 
				 PATH_BUF_SIZE,           // buffer size  
				 szName);                 // name of mapping object

	if (hMapFile == NULL) 
	{ 
	  DbgPrint("    [GWMain] Could not create file mapping object (%d)", GetLastError() );
	  return NULL;
	}

	*phSharedMem = hMapFile;

	pBuf = (LPTSTR) MapViewOfFile(hMapFile,   // handle to map object
						FILE_MAP_ALL_ACCESS,  // read/write permission
						0,                   
						0,                   
						PATH_BUF_SIZE);           
	if (pBuf == NULL) 
	{ 
	  DbgPrint("    [GWMain] Could not map view of file (%d)", GetLastError() ); 
	  CloseHandle(hMapFile);
	  return NULL;
	}

	return (PVOID)pBuf;
}


void CleanupSharedMemroy( LPVOID pMapBuf, HANDLE hShareMem )
{
	UnmapViewOfFile(pMapBuf);
	CloseHandle(hShareMem);
}


BOOL WritePath2SharedMemory( LPVOID lpPath, UINT ulLen )
{
	HANDLE hMapFile;
	LPCTSTR pBuf;

	hMapFile = OpenFileMapping(
				   FILE_MAP_ALL_ACCESS,   // read/write access
				   FALSE,                 // do not inherit the name
				   szName);               // name of mapping object 

	if (hMapFile == NULL) 
	{ 
		DbgPrint("  [GWMain] OS:Could not open file mapping object (%d)", GetLastError() );
		return FALSE;
	} 

	pBuf = (LPTSTR) MapViewOfFile(hMapFile, // handle to map object
			   FILE_MAP_ALL_ACCESS,			// read/write permission
			   0,                    
			   0,                    
			   PATH_BUF_SIZE);                   

	if (pBuf == NULL) 
	{ 
		DbgPrint("    [GWMain] OS:Could not map view of file (%d)", GetLastError() ); 

	  CloseHandle(hMapFile);

	  return FALSE;
	}

	// clear memory first
	ZeroMemory( (PVOID)pBuf, PATH_BUF_SIZE);
	// write the file path
	CopyMemory( (PVOID)pBuf, lpPath, ulLen);

	UnmapViewOfFile(pBuf);
	CloseHandle(hMapFile);
	
	return TRUE;
}


void SMWatchProc( LPVOID pParam )
{
	// copy parameter to local variable
	PATH_INFO* pPathInfo = (PATH_INFO*)pParam;

	CCurrentGroupInfo* pObjGroupInfo = (CCurrentGroupInfo*)(pPathInfo->pGroupInfoObj);
	HANDLE             hSMEvent      = pPathInfo->hSMEvent;

	// 1. create shared memroy
	HANDLE hSMFileHandle = 0;
	LPVOID pBuf          = CreateSharedMemory(&hSMFileHandle);

	if( NULL == pBuf )
	{
		return;
	}
	
	// 2. wait event
	DWORD dwRet = 0;
	
	while(1)
	{
		dwRet = ::WaitForSingleObject( hSMEvent, INFINITE );
		if( WAIT_OBJECT_0 != dwRet )
		{
			DbgPrint("    [SMWatchProc] SM event (err: %d)", GetLastError() );
		}
		// 3. if signaled, get file path and parse



		// 4. create a request and set signal
	}
}
#endif

BOOL WSAInit( void )
{
	WORD    wVersionRequested = MAKEWORD(2,2);
	WSADATA wsaData;
	BOOL    bRet = TRUE;

	/* initialize WSASocket sub system - Don't forget to call WSACleanup() */
	int nRet = WSAStartup(wVersionRequested, &wsaData);
	if( nRet != 0 )
	{
		bRet = FALSE;
	}
	
	return bRet;
}


void WSAShutdown( void )
{
	/* shutdownn winsock */
	WSACleanup();
}


void GetInterfaceAddress( string& strIP, DWORD* pdwIndex )
{
	/* get best local interface for multicast */
	IPAddr		stIPAddr;
	DWORD		dwBestIfIndex;
	DWORD		dwRet;

	stIPAddr = inet_addr("224.0.0.0");
	dwRet    = GetBestInterface( stIPAddr, &dwBestIfIndex );
	
	if( NO_ERROR == dwRet )
	{
		*pdwIndex = dwBestIfIndex;

		DbgPrint("[GWMain] Best interface index=:%d", dwBestIfIndex);
		
		/* Variables used by GetIpAddrTable */
		PMIB_IPADDRTABLE pIPAddrTable;
		DWORD dwSize   = 0;
		DWORD dwRetVal = 0;
		IN_ADDR IPAddr;
		int i = 0;

		/* Variables used to return error message */
		LPVOID lpMsgBuf;

		// Before calling AddIPAddress we use GetIpAddrTable to get
		// an adapter to which we can add the IP.
		pIPAddrTable = (MIB_IPADDRTABLE *) MALLOC(sizeof (MIB_IPADDRTABLE));

		if (pIPAddrTable) 
		{
			// Make an initial call to GetIpAddrTable to get the
			// necessary size into the dwSize variable
			if (GetIpAddrTable(pIPAddrTable, &dwSize, 0) == ERROR_INSUFFICIENT_BUFFER) 
			{
				FREE(pIPAddrTable);
				pIPAddrTable = (MIB_IPADDRTABLE *) MALLOC(dwSize);
			}
			if (pIPAddrTable == NULL) 
			{
				DbgPrint("    [GWMain] Memory allocation failed for GetIpAddrTable\n");
				return;
			}
		}
		// Make a second call to GetIpAddrTable to get the
		// actual data we want
		if ( (dwRetVal = GetIpAddrTable( pIPAddrTable, &dwSize, 0 )) != NO_ERROR ) 
		{ 
			DbgPrint("[GWMain] GetIpAddrTable failed with error %d\n", dwRetVal);
			if (FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, 
							  NULL, 
							  dwRetVal, 
							  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),  // Default language
							  (LPTSTR)& lpMsgBuf, 0, NULL) ) 
			{
				DbgPrint("    [GWMain] Error: %s", lpMsgBuf);
				LocalFree(lpMsgBuf);
			}
			return;
		}

		for (i=0; i < (int) pIPAddrTable->dwNumEntries; i++) 
		{
			if( pIPAddrTable->table[i].dwIndex == dwBestIfIndex )
			{
				IPAddr.S_un.S_addr = (u_long) pIPAddrTable->table[i].dwAddr;
		        strIP= inet_ntoa(IPAddr);
				DbgPrint("[GWMain] Best IF addr: %s", strIP.c_str());
				break;
			}
		}

		// dealloc memory
		if (pIPAddrTable) 
		{
			FREE(pIPAddrTable);
			pIPAddrTable = NULL;
		}
  
	}

}


BOOL GetInterfaceMACAddress( DWORD dwIndex, UCHAR* pMacBuffer )
{
	DWORD dwRet = 0;
	PIP_ADAPTER_INFO pAdapterInfo;
	ULONG ulOutBufLen;

	//ulOutBufLen = 20480;
	//char* pBuffer = new char[ulOutBufLen];

	ulOutBufLen = sizeof (IP_ADAPTER_INFO);
    pAdapterInfo = (IP_ADAPTER_INFO *) MALLOC(sizeof (IP_ADAPTER_INFO));
    if(pAdapterInfo == NULL) 
	{
        DbgPrint("    [GWMain]Error allocating memory needed to call GetAdaptersinfo");
        return FALSE;
    }
	// Make an initial call to GetAdaptersInfo to get
	// the necessary size into the ulOutBufLen variable
    if(GetAdaptersInfo(pAdapterInfo, &ulOutBufLen) == ERROR_BUFFER_OVERFLOW) 
	{
        FREE(pAdapterInfo);
		//printf("\n required meme: %d", ulOutBufLen );
        pAdapterInfo = (IP_ADAPTER_INFO *) MALLOC(ulOutBufLen);
        if (pAdapterInfo == NULL) 
		{
            printf("    [GWMain]Error allocating memory needed to call GetAdaptersinfo");
            return FALSE;
        }
    }

	dwRet = GetAdaptersInfo( pAdapterInfo, &ulOutBufLen);
	
	if( NO_ERROR == dwRet )
	{
		PIP_ADAPTER_INFO pInfoIter;
		pInfoIter = pAdapterInfo;
		
		while( pInfoIter != NULL )
		{
			//printf( "\nindex = %d, dwIndex = %d", pInfoIter->Index, dwIndex);
			if( dwIndex == pInfoIter->Index )
			{
				memcpy( pMacBuffer, pInfoIter->Address, pInfoIter->AddressLength );
				break;
			}
			pInfoIter = pInfoIter->Next;
		}
		
	}
	else
	{
		DbgPrint("    [GWMain] Fail to set MAC address");
		FREE(pAdapterInfo);
		return FALSE;
	}

	DbgPrint("[GWMain] MAC: %02x:%02x:%02x:%02x:%02x:%02x", 
		pMacBuffer[0], pMacBuffer[1], pMacBuffer[2], 
		pMacBuffer[3], pMacBuffer[4], pMacBuffer[5] );

	FREE(pAdapterInfo);

	return TRUE;
}


BOOL GetWMPlayerPath()
{
	//Read registery to find Windows media installed path


	LONG lRet = 0;
	HKEY hKey = NULL;
	DWORD dwOutSize = 0;

	// 1. open
	lRet = RegOpenKeyEx(
				HKEY_LOCAL_MACHINE,							// key
				TEXT("SOFTWARE\\Microsoft\\MediaPlayer"),	// subkey string
				0,
				KEY_READ,
				&hKey );                                    // handle
	
	if( lRet != ERROR_SUCCESS )
	{
		DbgPrint("    [GWMain] Fail to get WMPlayer path set default path");
		//strcpy_s( g_szWMPath, "C:\\Program Files\\Windows Media Player\\wmplayer.exe" );
		return FALSE;
	}


	// 2. query
	dwOutSize = sizeof(g_szWMPath);

	lRet = RegQueryValueEx(
				hKey,
				TEXT("Installation Directory"),	// value to find
				NULL,							// reserved
				NULL,							// out type
				(LPBYTE)g_szWMPath,				// out buffer
				&dwOutSize  );					// out len


	if( lRet != ERROR_SUCCESS )
	{
		DbgPrint("    [GWMain] Fail to get WMPlayer path set default path");
		//strcpy_s( g_szWMPath, "C:\\Program Files\\Windows Media Player\\wmplayer.exe" );
		RegCloseKey (hKey);
		return FALSE;
	}
	else
	{
		// add "wmplayer.exe"
		strcat_s(g_szWMPath, "\\wmplayer.exe");
		DbgPrint("[GWMain] WMPlayer path: %s", g_szWMPath );		
	}

	RegCloseKey (hKey);
	return TRUE;
}

BOOL GetVLCPlayerPath()
{
	//Read registery to find Windows media installed path


	LONG lRet = 0;
	HKEY hKey = NULL;
	DWORD dwOutSize = 0;

	// 1. open
	lRet = RegOpenKeyEx(
				HKEY_LOCAL_MACHINE,					// key
				TEXT("SOFTWARE\\VideoLAN\\VLC"),	// subkey string
				0,
				KEY_READ,
				&hKey );                            // handle
	
	if( lRet != ERROR_SUCCESS )
	{
		DbgPrint("    [GWMain] Fail to get VLC path");
		//strcpy_s( g_szWMPath, "C:\\Program Files\\Windows Media Player\\wmplayer.exe" );
		return FALSE;
	}


	// 2. query
	dwOutSize = sizeof(g_szWMPath);

	lRet = RegQueryValueEx(
				hKey,
				TEXT("InstallDir"),	// value to find
				NULL,							// reserved
				NULL,							// out type
				(LPBYTE)g_szVLCPath,			// out buffer
				&dwOutSize  );					// out len


	if( lRet != ERROR_SUCCESS )
	{
		DbgPrint("    [GWMain] Fail to get VLC path");
		//strcpy_s( g_szWMPath, "C:\\Program Files\\Windows Media Player\\wmplayer.exe" );
		RegCloseKey (hKey);
		return FALSE;
	}
	else
	{
		// add "wmplayer.exe"
		strcat_s(g_szVLCPath, "\\vlc.exe");
		DbgPrint("[GWMain] VLCPlayer path: %s", g_szVLCPath );		
	}

	RegCloseKey (hKey);
	return TRUE;
}


void WMPlayerRunProc( LPVOID pParam )
{
	//////////////////////////////////////////////////////////////////////////
	STARTUPINFO si;
	PROCESS_INFORMATION pi;
	CHAR szCommand[2048];
	//CHAR szExe[1024];
	//CHAR szFullExe[1024] = "\0";
	PCHAR pszFileExe = NULL;

	ZeroMemory(&si, sizeof(si));
	ZeroMemory(&pi, sizeof(pi));
	si.cb = sizeof(si);

	szCommand[0] = L'\0';

	// create command line
    //strcpy_s(szExe, sizeof(szExe), argv[arg]);
	//strcpy_s(szCommand, sizeof(szCommand), g_szWMPath);
	sprintf_s(szCommand, sizeof(szCommand),"\"%s\" \"%s\"", g_szWMPath, (char*)pParam );
		
	HWND hWnd = FindWindow( NULL, TEXT("Windows Media Player") );
	if( hWnd != NULL)
	{
		::SendMessage(hWnd, WM_CLOSE, 0, 0 );
		//::Sleep(500);
	}

	::Sleep(500);
	

	printf("\nStarting: `%s'\n", szCommand);
	//printf("withdll.exe:   with `%s'\n\n", szDllPath);
	//printf("withdll.exe:   marked by `%s'\n\n", szDetouredDllPath);
	//fflush(stdout);

    DWORD dwFlags = CREATE_DEFAULT_ERROR_MODE | CREATE_SUSPENDED;

    SetLastError(0);
    //SearchPath(NULL, szExe, ".exe", sizeof(szFullExe), szFullExe, &pszFileExe);
	BOOL bRet = FALSE;
	bRet = DetourCreateProcessWithDll(
				NULL, //szFullExe[0] ? szFullExe : NULL, 
				szCommand,
				NULL, 
				NULL, 
				FALSE,//TRUE, 
				dwFlags, 
				NULL, 
				NULL,
                &si, 
				&pi, 
				g_szDetourMarkerDLLPath,	// szDetouredDllPath, 
				g_szDetourDLLPath,			// szDllPath, 
				NULL);

    if( FALSE == bRet ) 
	{
        printf("Process creation failed: %d\n", GetLastError());
		return;
    }

    ResumeThread(pi.hThread);

    WaitForSingleObject(pi.hProcess, INFINITE);

    DWORD dwResult = 0;
    if( !GetExitCodeProcess(pi.hProcess, &dwResult) ) 
	{
        printf("Get exit code failed: %d\n", GetLastError());
		return;
    }
	
}

void VLCPlayerRunProc( LPVOID pParam )
{
	//////////////////////////////////////////////////////////////////////////
	STARTUPINFO si;
	PROCESS_INFORMATION pi;
	CHAR szCommand[2048];
	PCHAR pszFileExe = NULL;

	ZeroMemory(&si, sizeof(si));
	ZeroMemory(&pi, sizeof(pi));
	si.cb = sizeof(si);

	szCommand[0] = L'\0';

	sprintf_s(szCommand, sizeof(szCommand),"\"%s\" \"%s\"", g_szVLCPath, (char*)pParam );
#if 0		
	HWND hWnd = FindWindow( NULL, TEXT("Windows Media Player") );
	if( hWnd != NULL)
	{
		::SendMessage(hWnd, WM_CLOSE, 0, 0 );
		//::Sleep(500);
	}
#endif
	::Sleep(500);
	

	printf("\nStarting: `%s'\n", szCommand);

    DWORD dwFlags = CREATE_DEFAULT_ERROR_MODE | CREATE_SUSPENDED;
	//DWORD dwFlags = CREATE_SUSPENDED;

    SetLastError(0);

	BOOL bRet = FALSE;
	bRet = DetourCreateProcessWithDll(
				NULL, //szFullExe[0] ? szFullExe : NULL, 
				szCommand,
				NULL, 
				NULL, 
				FALSE,//TRUE, 
				dwFlags, 
				NULL, 
				NULL, 
                &si, 
				&pi, 
				g_szDetourMarkerDLLPath,	// szDetouredDllPath, 
				g_szDetourDLLPath,			// szDllPath, 
				NULL);

    if( FALSE == bRet ) 
	{
        printf("Process creation failed: %d\n", GetLastError());
		return;
    }

    ResumeThread(pi.hThread);

    WaitForSingleObject(pi.hProcess, INFINITE);

    DWORD dwResult = 0;
    if( !GetExitCodeProcess(pi.hProcess, &dwResult) ) 
	{
        printf("Get exit code failed: %d\n", GetLastError());
		return;
    }
	
}
/////////////////////////////////////////////////////////////
//
// Name: ParseInputFile
// Parameters: char*, null termicated input file path
//  string&, string object that will store player specific 
//  command line
// Return: int, player type defined in common.h file
// Descriptions:  This function parse the .amt input file and find
//   the player and its command line
//

int ParseInputFile( char* pszInFilePath, string& strCmd )
{
	int nPlayer = 0;
	////////////////////////////////////////////////////////////
	// Parse XML input file
	CXMLParserCtrl* pObjXMLParserCtrl = new CXMLParserCtrl( pszInFilePath );
	if( pObjXMLParserCtrl->LoadXML() == FALSE )
	{
		DbgPrint("    [GWMain] Failt to load the .amt file");
		return 0;
	}

	// start parsing
	pObjXMLParserCtrl->ParseAMTInfo();

	string strPlayer  = pObjXMLParserCtrl->GetPlayerName();
	strCmd = pObjXMLParserCtrl->GetCommandLine();

	if( pObjXMLParserCtrl != NULL )
		delete pObjXMLParserCtrl;

	if( strPlayer.compare("wmplayer") == 0 )
	{
		nPlayer = WM_PLAYER;
	}
	else if(strPlayer.compare("vlcplayer") == 0 )
	{
		nPlayer = VLC_PLAYER;
	}

	return nPlayer;
}


////////////////////////////////////////////////////////////////
//
// Name: SetPlayerPath
// Parameters: int, player type defined in the common.h file
// Return: BOOL, return true if there is path of a player, 
//   if not false
// Descriptions: the function find the path of the player
//

BOOL SetPlayerPath( int nPlayerType )
{
	////////////////////////////////////////////////////////////
	// Get player path - will be soted in g_szWMPath[]
	if( WM_PLAYER == nPlayerType )
	{
		if( NULL == g_szWMPath[0] )
		{
			if( FALSE == GetWMPlayerPath() )
			{
				DbgPrint("    [GWMain] Cannot find Windows Media Player");
				return FALSE;
			}
		}
	}
	else if( nPlayerType == VLC_PLAYER )
	{
		if( NULL == g_szVLCPath[0] )
		{

			if( FALSE == GetVLCPlayerPath() )
			{
				DbgPrint("    [GWMain] Cannot find VLC Player");
				return 0;
			}
		}
	}

	return TRUE;
}

//////////////////////////////////////////////////////////////////
//
//  Name: SendMutiInstanceMsg
//  Parameter: char*, input file path, null termincated string
//  Description: the function send the input file path to the
//    previously launched AMTGateway32 process
//
//
void SendMutiInstanceMsg( char* pszFilePath )
{

	SOCKET sCinet = socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP );

	SOCKADDR_IN stLocalAddr;
	SOCKADDR_IN stSvrAddr;

	// binding socket to local address
	stLocalAddr.sin_family      = AF_INET;
	stLocalAddr.sin_port        = htons(0);   
	stLocalAddr.sin_addr.s_addr = INADDR_ANY;

	int nRet = bind(
				sCinet,						// Socket descriptor
				(sockaddr*)&stLocalAddr,	// Address to bind to
				sizeof(struct sockaddr)	);	// Size of address
			

	// server aaddress
	stSvrAddr.sin_family      = AF_INET;
	stSvrAddr.sin_port        = htons(SVR_PORT);
	stSvrAddr.sin_addr.s_addr = inet_addr("127.0.0.1");

	int nLen = strlen(pszFilePath) + 1; // control code (1 byte) + input file path
	//char* pBuffer = new char[nLen];
	char pBuffer[2048];	
	ZeroMemory( pBuffer, sizeof(pBuffer) );

	pBuffer[0] = MULTI_INSTANCE;
	memcpy( &pBuffer[1], pszFilePath, nLen );
	
	//if( NULL != pBuffer )
	{
		nRet = sendto( 
				sCinet, 
				pBuffer, 
				nLen, 
				0, 
				(sockaddr*)&stSvrAddr, 
				sizeof(struct sockaddr) );

		//delete []  pBuffer;
	}
}

/////////////////////////////////////////////////////////////////
//
// Main 
//
// - initialize the player hook and parse command line
// - create AMT objects
// - get membership report messages and make request
// - launch appropriate player based on input file
//

int main(int argc, char **argv)
{
	int nPlayerType = 0;

	// clear memroy even though these are global
	ZeroMemory(g_szWMPath, sizeof(g_szWMPath) );
	ZeroMemory(g_szVLCPath, sizeof(g_szVLCPath) );
	ZeroMemory(g_szQTPath, sizeof(g_szQTPath) );

	//printf("argc[0] %s", argv[0] );
	//printf("argv[1] %s", argv[1] );
	//SendMutiInstanceMsg(argv[1]);


	//////////////////////////////////////////////////////////////
	// Initialize Winsock subsystem 
	if( FALSE == WSAInit() )
	{
		/* error - nothing to do, just exit */
		DbgPrint("    [GWMain] WSAStartup fail (err: %d)", WSAGetLastError() );
		return 0;
	}

	//////////////////////////////////////////////////////////////
	// check multiple instance
	if( HaveAnotherInstance() )
	{
		// send msaage to another instance
		SendMutiInstanceMsg(argv[1]);
		//::Sleep(1000);
		
		// cleanup winsock
		WSAShutdown();
		// exit silently
		return 0;
	}

	///////////////////////////////////////////////////////////////
	// Get netwrok interface address
	string		strIP;
	hostent*	pLocalHost;
	DWORD		dwIFIndex;
	UCHAR		auchMAC[6];

	ZeroMemory( auchMAC, sizeof(auchMAC) );

	GetInterfaceAddress(strIP, &dwIFIndex);

	if( strIP.empty() )
	{
		pLocalHost = gethostbyname("");
		strIP = inet_ntoa (*(struct in_addr *)*pLocalHost->h_addr_list);
	}

	if( FALSE == GetInterfaceMACAddress( dwIFIndex, auchMAC ) )
	//if( FALSE == GetInterfaceMACAddress( strIP, auchMAC ) )
	{
		DbgPrint("    [GWMain] Cannot set MAC address");
		return 0;
	}


	//////////////////////////////////////////////////////////////
	// Parse input file
	string strCommand;
	nPlayerType = ParseInputFile( argv[1], strCommand );
	if( 0 == nPlayerType )
	{
		// cannot do anything at all, just quit
		WSAShutdown();
		return 0;
	}
	if( FALSE == SetPlayerPath(nPlayerType) )
	{
		// cannot do anything at all, just quit
		WSAShutdown();
		return 0;
	}


	//////////////////////////////////////////////////////////////
	// prepare API hook

	// get detour dll path
	char szDetourDLL[] = "TraceMembership.dll";
	char szDeturedDLL[] = "detoured.dll";
	
	ZeroMemory( g_szDetourDLLPath, sizeof(g_szDetourDLLPath) );
	ZeroMemory( g_szDetourMarkerDLLPath, sizeof(g_szDetourMarkerDLLPath) );

	strcpy_s( g_szDetourDLLPath, argv[0] );
	strcpy_s( g_szDetourMarkerDLLPath, argv[0] );
	
	// so far my current directory info reside in the "szDetourDLLPath"
	char* pPos = NULL;
	pPos = strrchr( g_szDetourDLLPath, '\\' );
	if( pPos != NULL )
	{
		memcpy( pPos+1, szDetourDLL, sizeof(szDetourDLL) );
	}
	// copy detired marker dll name
	pPos = strrchr( g_szDetourMarkerDLLPath, '\\' );
	if( pPos != NULL )
	{
		memcpy( pPos+1, szDeturedDLL, sizeof(szDeturedDLL) );
	}
	

	

	/////////////////////////////////////////////////////////////
	// 
	// Create AMT handling objects
	//

	/* create instance of CAMTCtl class */
	CAMTCtrl* pObjAMTCtrl = new CAMTCtrl();
	if( NULL == pObjAMTCtrl )
	{
		DbgPrint("    [GWMain] Cannot create CAMTCtrl obj" );
		return 0;
	}

	/* create current group info */
	CCurrentGroupInfo* pObjGroupInfo = new CCurrentGroupInfo(); 
	if( NULL == pObjGroupInfo )
	{
		DbgPrint("    [GWMain] Cannot create CCurrentGroupInfo obj" );
		return 0;
	}
	pObjGroupInfo->SetIFAddress(strIP);


	/* create Pkt sender object */
	CPktSender* pObjPktSender = new CPktSender( strIP, auchMAC );
	if( NULL == pObjPktSender )
	{
		DbgPrint("    [GWMain] Cannot create CPktSender obj" );
		return 0;
	}
	
	/* pass pkt sender object to AMTCtrl */
	pObjAMTCtrl->SetPktSenderObject( pObjPktSender );
	

	CIGMPPktGen* pobjIGMPPktGen = new CIGMPPktGen();
	if( NULL == pobjIGMPPktGen )
	{
		DbgPrint("    [GWMain] Cannot create CIGMPPktGen obj" );
		return 0;
	}

	/////////////////////////////////////////////////////////////
	// Launch Player
	switch( nPlayerType )
	{
		case WM_PLAYER:
			_beginthread( WMPlayerRunProc, 0, (void*)strCommand.c_str() );
			break;
		case VLC_PLAYER:
			_beginthread( VLCPlayerRunProc, 0, (void*)strCommand.c_str() );
			break;
		case QT_PLAYER:
			break;
	}
	

	///////////////////////////////////////////////////////////////////
	// Main thread loop
	// 
	// 1. Parsing input file and extract (S,G)
	// 2. Make a request

	int nRcved = 0;
	int nSize  = 0;
	
	SOCKADDR_IN stAddrSvr;
	SOCKADDR_IN stAddrClient;

	char chRcvBuffer[2048];


	SOCKET sIPC = socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP );

	if( sIPC == INVALID_SOCKET )
	{
		DbgPrint("    [GWMain] Fail to initialize IPC socket");
		return 0;
	}

	// initialization for IPC socket
	ZeroMemory( &stAddrSvr, sizeof(SOCKADDR_IN) );

	stAddrSvr.sin_family      = AF_INET;
	stAddrSvr.sin_addr.s_addr = INADDR_ANY;
	stAddrSvr.sin_port        = htons(SVR_PORT);
	
	nSize = sizeof(SOCKADDR_IN);

	int nRet = bind(sIPC,(LPSOCKADDR)&stAddrSvr, nSize );
	if( nRet == SOCKET_ERROR )
	{
		DbgPrint("    [GWMain] Fail to bind IPC socket");
		return 0;
	}

	// for request generation
	unsigned char* pOutBuffer;
	int nOutSize;

	while(1)
	{
		pOutBuffer = NULL;
		nOutSize   = 0;

		// clear memroy before receive the message
		ZeroMemory( chRcvBuffer, sizeof(chRcvBuffer) );
		
		// receive
		nRcved = recvfrom( 
					sIPC, 
					chRcvBuffer, 
					sizeof(chRcvBuffer), 
					0, 
					(sockaddr*)&stAddrClient, &nSize );
		
		// check error
		if( nRcved == SOCKET_ERROR )
		{
			int nErrCode = WSAGetLastError();
			DbgPrint("    [GWMain] Fail to receive (err:%d)", nErrCode );
			// go back to while loop
			continue;
		}

		// parse message
		// +---------------------------------------- 
		// | Type | Message
		// +----------------------------------------
		switch( (int)chRcvBuffer[0] )
		{
			case MEMBERSHIP_UPDATE:
				{
					// message: setsockopt option name (int) + option val (variable)
					int nOption = 0;
					memcpy( &nOption, &chRcvBuffer[1], sizeof(int) );

					IP_MREQ_SOURCE	stMEMBESHIP;
					IP_MREQ			stMREQ;

					ZeroMemory( &stMREQ, sizeof(IP_MREQ) );
					ZeroMemory( &stMEMBESHIP, sizeof(IP_MREQ_SOURCE) );
					

					switch( nOption )
					{
						case IP_ADD_SOURCE_MEMBERSHIP:
							{

								DbgPrint("IP_ADD_SOURCE_MEMBERSHIP");
								memcpy( &stMEMBESHIP, &chRcvBuffer[1]+sizeof(int), sizeof(IP_MREQ_SOURCE) );
								//printf( "S:%s, G:%s\n", inet_ntoa( stMEMBESHIP.imr_sourceaddr), 
								//					    inet_ntoa( stMEMBESHIP.imr_multiaddr) );
								DbgPrint( "-----------------------------------------------");
								DbgPrint( "S:%s", inet_ntoa( stMEMBESHIP.imr_sourceaddr ) );
								DbgPrint( "G:%s", inet_ntoa( stMEMBESHIP.imr_multiaddr ) );
								DbgPrint( "-----------------------------------------------");

								BOOL bSuccess = pObjGroupInfo->AddGroup(
													stMEMBESHIP.imr_multiaddr.S_un.S_addr,
													stMEMBESHIP.imr_sourceaddr.S_un.S_addr );
								
								if( FALSE == bSuccess )
									break; //duplicate item

								// make a IGMP packek for membership join
								ULONG aulSrc[1];
								aulSrc[0]   = stMEMBESHIP.imr_sourceaddr.S_un.S_addr;
								ULONG ulGrp = stMEMBESHIP.imr_multiaddr.S_un.S_addr;
								
								pobjIGMPPktGen->GetIGMPMembershipPkt( 
									TRUE, aulSrc, 1, ulGrp, strIP.c_str(), &pOutBuffer, &nOutSize );

								// create a request &  put it in the request queue
								if( pOutBuffer != NULL && nOutSize != 0)
								{
									REQUEST newRequest;
									ZeroMemory( &newRequest, sizeof(REQUEST) );

									newRequest.pucIGMPData  = pOutBuffer;
									newRequest.ulIGMPLength = nOutSize;

									pObjAMTCtrl->AddRequest( newRequest );
								}
							}
							break;

						case MCAST_JOIN_SOURCE_GROUP:
							{
								DbgPrint("MCAST_JOIN_SOURCE_GROUP");
								//GROUP_SOURCE_REQ stGRP_SRC_REQ;
								struct group_source_req stGRP_SRC_REQ;
								memcpy( &stGRP_SRC_REQ, &chRcvBuffer[1]+sizeof(int), sizeof(struct group_source_req) );
								//printf( "S:%s, G:%s\n", inet_ntoa( stMEMBESHIP.imr_sourceaddr), 
								//					    inet_ntoa( stMEMBESHIP.imr_multiaddr) );

								SOCKADDR_IN stSRC, stGRP;
								memcpy(&stSRC, &stGRP_SRC_REQ.gsr_source, sizeof(SOCKADDR) );
								memcpy(&stGRP, &stGRP_SRC_REQ.gsr_group, sizeof(SOCKADDR) );


								DbgPrint( "-----------------------------------------------");
								DbgPrint( "S:%s", inet_ntoa( stSRC.sin_addr ) );
								DbgPrint( "G:%s", inet_ntoa( stGRP.sin_addr ) );
								DbgPrint( "-----------------------------------------------");

								BOOL bSuccess = pObjGroupInfo->AddGroup(
													stGRP.sin_addr.S_un.S_addr,
													stSRC.sin_addr.S_un.S_addr);
								
								if( FALSE == bSuccess )
									break; //duplicate item

								// make a IGMP packek for membership join
								ULONG aulSrc[1];
								//aulSrc[0]   = stMEMBESHIP.imr_sourceaddr.S_un.S_addr;
								//ULONG ulGrp = stMEMBESHIP.imr_multiaddr.S_un.S_addr;
								aulSrc[0]   = stSRC.sin_addr.S_un.S_addr;
								ULONG ulGrp = stGRP.sin_addr.S_un.S_addr;
								
								pobjIGMPPktGen->GetIGMPMembershipPkt( 
									TRUE, aulSrc, 1, ulGrp, strIP.c_str(), &pOutBuffer, &nOutSize );

								// create a request &  put it in the request queue
								if( pOutBuffer != NULL && nOutSize != 0)
								{
									REQUEST newRequest;
									ZeroMemory( &newRequest, sizeof(REQUEST) );

									newRequest.pucIGMPData  = pOutBuffer;
									newRequest.ulIGMPLength = nOutSize;

									pObjAMTCtrl->AddRequest( newRequest );
								}

							}
							break;

						case IP_DROP_SOURCE_MEMBERSHIP:
							{
								DbgPrint("IP_DROP_SOURCE_MEMBERSHIP");
								memcpy( &stMEMBESHIP, &chRcvBuffer[1]+sizeof(int), sizeof(IP_MREQ_SOURCE) );
								//printf( "S:%s, G:%s\n", inet_ntoa( stMEMBESHIP.imr_sourceaddr), 
								//					    inet_ntoa( stMEMBESHIP.imr_multiaddr) );
								DbgPrint( "-----------------------------------------------");
								DbgPrint( "S:%s", inet_ntoa( stMEMBESHIP.imr_sourceaddr ) );
								DbgPrint( "G:%s", inet_ntoa( stMEMBESHIP.imr_multiaddr ) );
								DbgPrint( "-----------------------------------------------");

								BOOL bSuccess = pObjGroupInfo->DeleteGroup(
													stMEMBESHIP.imr_multiaddr.S_un.S_addr,
													stMEMBESHIP.imr_sourceaddr.S_un.S_addr );
								if( FALSE == bSuccess )
									break; // No itme to remove

								// make a IGMP packek for membership leave
								ULONG aulSrc[1];
								aulSrc[0]   = stMEMBESHIP.imr_sourceaddr.S_un.S_addr;
								ULONG ulGrp = stMEMBESHIP.imr_multiaddr.S_un.S_addr;
								
								pobjIGMPPktGen->GetIGMPMembershipPkt( 
									FALSE, aulSrc, 1, ulGrp, strIP.c_str(), &pOutBuffer, &nOutSize );
								
								// create a request &  put it in the request queue
								if( pOutBuffer != NULL && nOutSize != 0)
								{
									REQUEST newRequest;
									ZeroMemory( &newRequest, sizeof(REQUEST) );

									newRequest.pucIGMPData  = pOutBuffer;
									newRequest.ulIGMPLength = nOutSize;

									pObjAMTCtrl->AddRequest( newRequest );
								}
							}
							break;

						default:
							break;
					}// end of inner switch
				}
				break; // end of MEMBERSHIP_UPDATE case

			case PERIODIC_QUERY:
				{
					int nSrcSize = 0;
					ULONG ulGrpAddr = 0;
					int nGrpSize = pObjGroupInfo->GetSize();

					if( nGrpSize == 0 )
						break;

					for( int i = 0; i < nGrpSize; i++ )
					{
						nSrcSize = pObjGroupInfo->GetSourceSize(i);
						ULONG* pSrcArray = new ULONG[nSrcSize];
						ulGrpAddr = pObjGroupInfo->GetSrc(i, pSrcArray);

						//get igmp pkt
						pobjIGMPPktGen->GetIGMPMembershipPkt( 
									TRUE, pSrcArray, nSrcSize, ulGrpAddr, strIP.c_str(), &pOutBuffer, &nOutSize );

						// create a request &  put it in the request queue
						if( pOutBuffer != NULL && nOutSize != 0)
						{
							REQUEST newRequest;
							ZeroMemory( &newRequest, sizeof(REQUEST) );

							newRequest.pucIGMPData  = pOutBuffer;
							newRequest.ulIGMPLength = nOutSize;

							pObjAMTCtrl->AddRequest( newRequest );
						}

						pOutBuffer = NULL;
						nOutSize = 0;

						delete [] pSrcArray;
					} // end of for loop
					
				}
				break;

			case MULTI_INSTANCE:
				{
					// 1. parsing the input file
					string strFile("\"");
					strCommand.clear(); // declared in main() scope
					
					strFile.append( &chRcvBuffer[1] );
					strFile.append("\"");
					
					nPlayerType = ParseInputFile( &chRcvBuffer[1], strCommand );
					//nPlayerType = ParseInputFile( &szBuffer[1], strCommand );
					printf("input: %s", strCommand.c_str() );

					if( 0 != nPlayerType )
					{
						SetPlayerPath(nPlayerType);
						// 2. launch a player
						if( WM_PLAYER == nPlayerType && FALSE == strCommand.empty() )
						{
							_beginthread( WMPlayerRunProc, 0, (LPVOID)strCommand.c_str() );
						}
						else if( VLC_PLAYER == nPlayerType && FALSE == strCommand.empty() )
						{
							_beginthread( VLCPlayerRunProc, 0, (LPVOID)strCommand.c_str() );
						}
					}

				}
				break;

			case 0x90:
				printf("Msg: %s\n", &chRcvBuffer[1] );
				break;

			default:
				break;
		}// end of outer switch

	}


	///////////////////////////////////////////////
	// delete objects 
	// Since its console application, it will never have a chance to cleanup the resources.
	// However, it's a good practice. ;-)
	// 
	if( NULL != pObjGroupInfo )
		delete pObjGroupInfo;
	if( NULL != pObjAMTCtrl )
		delete pObjAMTCtrl;
	if( NULL != pObjPktSender )
		delete pObjPktSender;

	WSAShutdown();
	

	return 0;
}

#if 0  // only for the detour dll test
						case IP_ADD_MEMBERSHIP:
							
							printf("IP_ADD_MEMBERSHIP\n");
							memcpy( &stMREQ, &chRcvBuffer[1]+sizeof(int), sizeof(IP_MREQ) );
							//printf( "I:%s, G:%s\n", inet_ntoa( stMREQ.imr_interface), 
							//	inet_ntoa( stMREQ.imr_multiaddr) );
							printf( "-----------------------------------------------\n");
							printf( "I:%s\n", inet_ntoa( stMREQ.imr_interface) );
							printf( "G:%s\n", inet_ntoa( stMREQ.imr_multiaddr) );
							printf( "-----------------------------------------------\n");
								
							break;

						case IP_DROP_MEMBERSHIP:
							printf("IP_DROP_MEMBERSHIP\n");
							memcpy( &stMREQ, &chRcvBuffer[1]+sizeof(int), sizeof(IP_MREQ) );
							//printf( "I:%s, G:%s\n", inet_ntoa( stMREQ.imr_interface), 
							//					    inet_ntoa( stMREQ.imr_multiaddr) );
							printf( "-----------------------------------------------\n");
							printf( "I:%s\n", inet_ntoa( stMREQ.imr_interface) );
							printf( "G:%s\n", inet_ntoa( stMREQ.imr_multiaddr) );
							printf( "-----------------------------------------------\n");
							break;

						case IP_BLOCK_SOURCE:
							printf("IP_BLOCK_SOURCE\n");
							memcpy( &stMEMBESHIP, &chRcvBuffer[1]+sizeof(int), sizeof(IP_MREQ_SOURCE) );
							//printf( "S:%s, G:%s\n", inet_ntoa( stMEMBESHIP.imr_sourceaddr), 
							//					    inet_ntoa( stMEMBESHIP.imr_multiaddr) );
							printf( "-----------------------------------------------\n");
							printf( "S:%s\n", inet_ntoa( stMEMBESHIP.imr_sourceaddr ) );
							printf( "G:%s\n", inet_ntoa( stMEMBESHIP.imr_multiaddr ) );
							printf( "-----------------------------------------------\n");
							break;

						case IP_UNBLOCK_SOURCE:
							printf("IP_UNBLOCK_SOURCE\n");
							memcpy( &stMEMBESHIP, &chRcvBuffer[1]+sizeof(int), sizeof(IP_MREQ_SOURCE) );
							//printf( "S:%s, G:%s\n", inet_ntoa( stMEMBESHIP.imr_sourceaddr), 
							//					    inet_ntoa( stMEMBESHIP.imr_multiaddr) );
							printf( "-----------------------------------------------\n");
							printf( "S:%s\n", inet_ntoa( stMEMBESHIP.imr_sourceaddr ) );
							printf( "G:%s\n", inet_ntoa( stMEMBESHIP.imr_multiaddr ) );
							printf( "-----------------------------------------------\n");
							break;


#endif
