#include "IGMPPktGen.h"

CIGMPPktGen::CIGMPPktGen(void)
{
}

CIGMPPktGen::~CIGMPPktGen(void)
{
}

#if 1
void CIGMPPktGen::GetIGMPMembershipPkt( BOOL bMode, 
										ULONG* pulSrc,
										int nNumSrc,
										ULONG ulGroup,
										const char* szSrcIP,
										unsigned char** ppOutBuffer,
										int* pnOutLen )
{

	int nRecSize  = sizeof(GROUP_RECORD) + ( sizeof(ULONG) * (nNumSrc-1) ) ;
	int nIGMPSize = sizeof(IGMPV3_MEMBERSHIP) + ( sizeof(ULONG) * (nNumSrc-1) ); // struct already include one record
	int nPktSize  = SIZE_IPHEADER + nIGMPSize;

	PGROUP_RECORD		pstGrpRec;
	PIGMPV3_MEMBERSHIP	pstIGMPMembership;

	// don't forget to delete memory
	unsigned char* pGrpRecBuffer = new unsigned char[nRecSize];
	unsigned char* pIGMPBuffer   = new unsigned char[nIGMPSize];
	
	if( NULL == pGrpRecBuffer || NULL == pIGMPBuffer )
	{
		*ppOutBuffer = NULL;
		*pnOutLen = 0;
		return;
	}

	ZeroMemory( pGrpRecBuffer, nRecSize );
	ZeroMemory( pIGMPBuffer, nIGMPSize );

	// map temporary buffer to allocated memory
	pstGrpRec         = (PGROUP_RECORD)pGrpRecBuffer;
	pstIGMPMembership = (PIGMPV3_MEMBERSHIP)pIGMPBuffer;

	
	// constuct grop record first
	//#define IGMPV3_MODE_IS_INCLUDE				(1)
	//#define IGMPV3_MODE_IS_EXCLUDE				(2)
	pstGrpRec->uchRecType    = bMode ? 5:6; 
	pstGrpRec->uchAuxLen     = 0;
	pstGrpRec->usNumSrc      = htons(1);
	pstGrpRec->ulGrpAddr     = ulGroup;
	for( int i = 0; i < nNumSrc; i++ )
	{
		pstGrpRec->aulSrcAddr[i] = pulSrc[i];
	}
	
	pstIGMPMembership->uchType        = 0x22; // membership report
	pstIGMPMembership->usChecksum     = 0;
	pstIGMPMembership->usNumGroup     = htons(1);
	memcpy( &(pstIGMPMembership->stGrpRecord[0]), pstGrpRec, nRecSize);

	// get checksum
	pstIGMPMembership->usChecksum = 
		GetChecksum( (unsigned short*)pIGMPBuffer, nIGMPSize );


	//////////////////////////////////////////////////////////////////////
	// assemble IP header

	// memory allocated in here will be deallocated at AMT request handling procedure
	unsigned char* pIGMPPkt = new unsigned char[nPktSize];
	
	
	IP_HEADER	stIP;

	stIP.uchVerLen   = 0x45; // IP4, length 20 byte
	stIP.uchTypeSvc  = 0x00;
	stIP.usTotLen    = htons(nPktSize); // IP header + IGMP length
	stIP.usID        = htons(0x11E1);
	stIP.usFlagFrag  = 0x00;
	stIP.uchTTL      = 0x01;
	stIP.uchProtocol = 0x02;
	stIP.usChecksum  = 0x00;
	stIP.ulSrcIP     = inet_addr( szSrcIP );
	stIP.ulDestIP    = inet_addr( ALL_IGMPV3_ROUTERS );

	// get IP checksum
	stIP.usChecksum = GetChecksum( (unsigned short*)&stIP, SIZE_IPHEADER );

	// copy ip header to buffer
	memcpy( pIGMPPkt, &stIP, SIZE_IPHEADER );

	// copu igmp to buffer
	memcpy( pIGMPPkt+SIZE_IPHEADER, pstIGMPMembership, nIGMPSize );

	// deallocate temporary memory
	if( NULL != pGrpRecBuffer )
	{
		delete [] pGrpRecBuffer;
	}

	if( NULL != pstIGMPMembership )
	{
		delete [] pstIGMPMembership;
	}

	*ppOutBuffer = pIGMPPkt;
	*pnOutLen   = nPktSize;
}


#endif

unsigned short CIGMPPktGen::GetChecksum(unsigned short *pusIPBuffer, int nLen)
{
	int nleft = nLen;
    int sum = 0;
    unsigned short *w = pusIPBuffer;
    unsigned short answer = 0;

    while (nleft > 1) 
	{
        sum += *w++;
        nleft -= 2;
    }
    if (nleft == 1) 
	{
        *(unsigned char*)(&answer) = *(unsigned char*)w;
        sum += answer;
    }
    sum = (sum >> 16) + (sum & 0xffff);
    answer = ~sum;
    return (answer);
}
