/*
* AMT Gateway Win32 version
* 
* File: PktMonitor.cpp
*
* Description: watching local IGMP packet and create request
*
* THE UNIVERSITY OF TEXAS AT DALLAS, ANDES Lab.
* Juhyun Oh (Justin), justinoh@student.utdallas.edu
*
*/


#include "PktMonitor.h"

CRITICAL_SECTION CPktMonitor::g_csTempReqQ;

CPktMonitor::CPktMonitor(void)
{
	//InitializePktMonProc();
	DbgPrint("    [PktMon] you should create a signal handle and pass it to this class");

}

CPktMonitor::CPktMonitor(HANDLE hReRcv )
{
	InitializeCriticalSection(&g_csTempReqQ);

	/* copy handle for signaling */
	if( NULL != hReRcv )
	{
		m_hReqRcvEvent = hReRcv;
	}

	/* initilaize mac address space */
	//ZeroMemory(&m_acLocalMAC, sizeof(m_acLocalMAC) );
	m_bRefreshMAC = TRUE;
}


CPktMonitor::~CPktMonitor(void)
{
	//DeleteCriticalSection(&g_csTempReqQ);
}

void CPktMonitor::InitializePktMonProc( void )
{
	/* create thread and check returned handle */
	m_hPktMonThread = (HANDLE)_beginthread( this->PktMonProc, 0, (void*)this );
	
	if( -1L ==  (long)m_hPktMonThread )
	{
		DbgPrint("    [PktMon] packet monitor thread thread creation failed" );
	}
}


/*
*
*  set current active interface address from user of this class
*  it will be used to find out the network interface, which has the same ip address
*/
void CPktMonitor::SetInterfaceIP(string & strIPAddress)
{
	if( strIPAddress.size() > 0 )
	{
		m_strUsrSetInterfaceIP = strIPAddress;
		DbgPrint("[PktMon] Provided IP address:%s", m_strUsrSetInterfaceIP.c_str());
	}
	else
	{
		DbgPrint("    [PktMon] Provided active interface IP is NULL");
	}
}

/* From tcptraceroute, convert a numeric IP address to a string */
char* CPktMonitor::iptos(unsigned long in)
{
    static char output[IPTOSBUFFERS][3*4+3+1];
    static short which;
    u_char *p;

    p = (u_char *)&in;
    which = (which + 1 == IPTOSBUFFERS ? 0 : which + 1);
    sprintf_s(output[which], "%d.%d.%d.%d", p[0], p[1], p[2], p[3]);
    return output[which];
}


BOOL CPktMonitor::InitializePcap( string& strActiveIFAddr )
{
	pcap_if_t*    pDevList;
	pcap_if_t*    pDevIter;
	pcap_addr_t*  pAddress;
	char          errbuf[PCAP_ERRBUF_SIZE+1];
	char          szSource[] = {"rpcap://"};
	string        strIFAddress;
	
	
	pDevList = NULL;
	pDevIter = NULL;


	/* set active interface IP */
	SetInterfaceIP(strActiveIFAddr);

	/* find all available interfaces and find active interface */
	if (pcap_findalldevs_ex(szSource, NULL, &pDevList, errbuf) == -1)
	{
		DbgPrint("    [PktMon] pcap_findalldevs_ex() fail: %s", errbuf );
		return FALSE;
	}


	 /* IP addresses */
	for( pDevIter = pDevList; pDevIter ; pDevIter = pDevIter->next )
	{
		//printf("\tAddress Family: #%d\n", pAddress->addr->sa_family);
		for( pAddress = pDevIter->addresses; pAddress ; pAddress = pAddress->next )
		{

			/* only for the not null item */
			if( NULL != pAddress->addr )
			{
				strIFAddress = iptos( ((struct sockaddr_in *)pAddress->addr)->sin_addr.s_addr );

				if( 0 == strIFAddress.compare(m_strUsrSetInterfaceIP) )
				{
					DbgPrint("[PktMon] IF selection info");
					DbgPrint("[PktMon] Name: %s", pDevIter->name);
					DbgPrint("[PktMon] Description: %s", pDevIter->description);
					
					/* copy interface name */
					m_strActiveIF = (pDevIter->name) + sizeof(szSource) - 1;
					pcap_freealldevs(pDevList);

					return TRUE;
				}

			}
		} // end of inner for loop

	}// end of for loop

	/* dealloc memory */
	pcap_freealldevs(pDevList);

	return FALSE;
}


string& CPktMonitor::GetActiveIFName()
{
	return m_strActiveIF;
}
/*
*
*/

int CPktMonitor::GetRequest( REQUEST& newReq )
{
	int nSize = 0;

	EnterCriticalSection(&g_csTempReqQ);

	nSize = m_listTempReqQ.size();
	DbgPrint("[PktMon] tempReqQ size: %d", nSize );

	if( nSize > 0 )
	{
		/* copy */
		newReq = m_listTempReqQ.front();

		/* remove */
		m_listTempReqQ.pop_front();

		nSize = m_listTempReqQ.size();
		//printf("\n [PktMon] (2) tempReqQ size: %d", nSize );
	}
	else
	{
		DbgPrint("    [PktMon] Req Q signaled, but it is empty");
	}

	LeaveCriticalSection(&g_csTempReqQ);

	return nSize;
	
}

void CPktMonitor::SetPktSenderObj( CPktSender* pParma )
{
	m_pObjPkrtSender = pParma;
}

void CPktMonitor::SetGroupInfoObj( CCurrentGroupInfo* pParma )
{
	m_pObjGroupInfo = pParma;
}

/*
*
*/
void CPktMonitor::PktMonProc( void* pParam )
{

	CPktMonitor*    pParent;
	string          strActiveIF;
	
	struct bpf_program fcode;
	bpf_u_int32     NetMask;
	
	char errbuf[PCAP_ERRBUF_SIZE];
	char szFilter[] = {"igmp"};

	int res;
	struct pcap_pkthdr*    pPktCaptureHeader;
	const u_char *pkt_data;

	
	pParent     = (CPktMonitor*)pParam;
	strActiveIF = pParent->m_strActiveIF;

	
	/* open a capture from the network */
	if( 0 < strActiveIF.size() )
	{
		pParent->m_pPcapDevHndl = pcap_open_live( strActiveIF.c_str() , // name of the device
			                                      MAX_PKT_CAPTURE_SIZE, // portion of the packet to capture. 65536 grants that the whole packet will be captured on all the MACs.
			                                      PCAP_OPENFLAG_NOCAPTURE_LOCAL ,//0,                    // promiscuous mode (nonzero means promiscuous)
			                                      1000,                 // read timeout
			                                      errbuf                // error buffer
			                                      );
		if( NULL == pParent->m_pPcapDevHndl )
		{
			DbgPrint("	[PktMon] Unable to oprn the adapter");
			::MessageBox(NULL, TEXT("pcap - Unable to oprn the adapter"), NULL, MB_OK | MB_ICONEXCLAMATION );
			return;
		}
	}
	//else usage();

	if ( NULL != szFilter )
	{
		// Let's do things simpler: we suppose to be in a C class network ;-)
		NetMask=0xffffff;

		//compile the filter
		if( pcap_compile(pParent->m_pPcapDevHndl, &fcode, szFilter, 1, NetMask) < 0)
		{
			pcap_close(pParent->m_pPcapDevHndl);
			return;
		}

		//set the filter
		if(pcap_setfilter(pParent->m_pPcapDevHndl, &fcode)<0)
		{
			pcap_close(pParent->m_pPcapDevHndl);
			return;
		}

	}

	REQUEST       newReq;
	string        strSourceIP;
	char          achTempBuffer[20];
	unsigned int  unIPHeaderLen = 0; 
	unsigned int  unIGMPOffset = 0;  //igmp base offset in captured packet

	//start the capture
 	while((res = pcap_next_ex( pParent->m_pPcapDevHndl, &pPktCaptureHeader, &pkt_data)) >= 0)
	{
		
		
		/* Timeout elapsed */
		if( 0 == res )
		{
			continue;
		}
		
		ZeroMemory(achTempBuffer, sizeof(achTempBuffer) );
		sprintf_s(achTempBuffer, "%d.%d.%d.%d", (int)pkt_data[PCAP_IPSRC_OFFSET],
			                                    (int)pkt_data[PCAP_IPSRC_OFFSET+1],
												(int)pkt_data[PCAP_IPSRC_OFFSET+2],
												(int)pkt_data[PCAP_IPSRC_OFFSET+3]);

		strSourceIP = achTempBuffer;		
		//cout<<"[PktMon] Captured IGMP source IP: "<<strSourceIP<<endl;

		if(  0 != strSourceIP.compare(pParent->m_strUsrSetInterfaceIP) )
		{
			//cout<<"[PktMon] Skip IGMP request not from local host"<<endl;
			continue;
		}

		/* filter out IGMP query which is generated by myself */
		unIPHeaderLen = (pkt_data[PCAP_IPHEADER_OFFSET] & 0x0F) * 4;
		unIGMPOffset  = PCAP_IPHEADER_OFFSET + unIPHeaderLen;

		if( 0x11 == pkt_data[unIGMPOffset] )
		{
			DbgPrint("[PktMon] Filter out general query from myself");
			continue;
		}

		/* filter out IGMP from the administrative multicast address */
		// 239.255.x.y
		if( 239 == (int)pkt_data[PCAP_IPDEST_OFFSET] &&
			255 == (int)pkt_data[PCAP_IPDEST_OFFSET+1] )
		{
			DbgPrint("[PktMon] Filter out admin address IGMP");
			continue;
		}


		/* check IGMP history - suppress short term duplicate IGMPs */
		
		//1. get current time
		//time_t stCurrentTime, stTimeDiff;
		//stCurrentTime = time(NULL);
		unsigned int unTimeDiff = 0;
		unsigned int nIGMPLen   = 0;
		BOOL         bDupIGMP   = FALSE;
		
		// get igmp packet length
		nIGMPLen = pPktCaptureHeader->caplen - unIGMPOffset;
		

		list<LATESTIGMP>::iterator	iter;

		iter = pParent->m_listIGMPHistory.begin();

		while( iter != pParent->m_listIGMPHistory.end() )
		{
			// 2. check timestmap first
			unTimeDiff = (unsigned int)pPktCaptureHeader->ts.tv_sec - ((LATESTIGMP)*iter).unTimestamp;

			if( unTimeDiff > IGMP_INTERVAL )
			{
				// 3. remove old IGMP stored more than IGMP_INTERVAL
				delete [] ((LATESTIGMP)*iter).pIGMPPkt;
				// remove current item and iterator moves to next item
				iter = pParent->m_listIGMPHistory.erase(iter);
				continue;
			}
			else
			{
				// 4. Is it the same IGMP?
				unsigned char* pucTemp = ((LATESTIGMP)*iter).pIGMPPkt;
				if( memcmp( &pkt_data[unIGMPOffset], pucTemp, nIGMPLen ) == 0 )
				{
					bDupIGMP = TRUE;
					break;
				}

				iter++;
			}
		}

		/* update IGMP history */
		if( FALSE == bDupIGMP )
		{
			LATESTIGMP stNewIGMP;
			stNewIGMP.pIGMPPkt = new unsigned char[nIGMPLen];
			memcpy( stNewIGMP.pIGMPPkt, &pkt_data[unIGMPOffset], nIGMPLen );
			stNewIGMP.unTimestamp = (unsigned int)pPktCaptureHeader->ts.tv_sec;

			pParent->m_listIGMPHistory.push_back(stNewIGMP);
		}
		else
		{
			/* skip it */
			continue;
		}

#if 0
		/* fill current group list - only look at IGMP portion of captured packet */
		pParent->m_pObjGroupInfo->UpdateGroupInfo( (unsigned char*)&pkt_data[unIGMPOffset],  nIGMPLen);
#endif

		
		if( TRUE == pParent->m_bRefreshMAC )
		{
			/* copy MAC addres of active interface */
			//memcpy( &(pParent->m_acLocalMAC), &pkt_data[6], 6 );
			pParent->m_bRefreshMAC = FALSE;
			pParent->m_pObjPkrtSender->SetMACLocalMACAddr((unsigned char*)&pkt_data[6]);
		}

#if 0
		DbgPrint( "[PktMon] Capture Length: %d", pPktCaptureHeader->caplen);
		DbgPrint( "0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x", pkt_data[0], pkt_data[1], pkt_data[2], pkt_data[3],
			                                                 pkt_data[4], pkt_data[5], pkt_data[6], pkt_data[7]);
		DbgPrint( "0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x", pkt_data[8], pkt_data[9], pkt_data[10], pkt_data[11],
			                                                 pkt_data[12], pkt_data[13], pkt_data[14], pkt_data[15]);
		DbgPrint( "0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x", pkt_data[16], pkt_data[17], pkt_data[18], pkt_data[19],
			                                                 pkt_data[20], pkt_data[21], pkt_data[22], pkt_data[23]);

#endif
		/* initialize reqest struct */
		ZeroMemory( &newReq, sizeof(REQUEST) );

		/* set length and allocate memory */
		// the IGMP includes IP HEADER
		newReq.ulIGMPLength = pPktCaptureHeader->caplen - PCAP_IPHEADER_OFFSET;
		newReq.pucIGMPData  = new unsigned char[newReq.ulIGMPLength];

		if( NULL != newReq.pucIGMPData )
		{
#ifdef MEM_DEBUG
			char szMessage[1024];
			ZeroMemory( szMessage, sizeof(szMessage) );
			int nSize = sprintf_s(szMessage, "0x%08x@File: %s, Line:%d", newReq.pucIGMPData, __FILE__, __LINE__);
			Mem_Debug(szMessage, nSize, TRUE );
#endif

#if 0 // temporary block
			///printf("\n[+]IGMP_REQ: 0x%x",newReq.pucIGMPData);
			memcpy( newReq.pucIGMPData, &pkt_data[PCAP_IPHEADER_OFFSET], newReq.ulIGMPLength );
			
			EnterCriticalSection(&g_csTempReqQ);
			pParent->m_listTempReqQ.push_back(newReq);
			LeaveCriticalSection(&g_csTempReqQ);

			/* save it for making new request */
			::SetEvent(pParent->m_hReqRcvEvent);
#endif
		}
		else
		{
			/* what shall I do??*/
			DbgPrint("    [PktMon] Fail to allocate request memory");
		}
	}
	pcap_close(pParent->m_pPcapDevHndl);
}


