/*
* AMT Gateway Win32 version
* 
* File: PktSender.cpp
*
* Description: send packets to the kernel by using pcap
*
* THE UNIVERSITY OF TEXAS AT DALLAS, ANDES Lab.
* Juhyun Oh (Justin), justinoh@student.utdallas.edu
*
*/


#include "PktSender.h"


CPktSender::CPktSender(void)
{
}

CPktSender::CPktSender(string& strParam, unsigned char* pMACBuffer)
{
	BOOL bRet;

	if( strParam.size() > 0 )
	{
		// copy MAC address
		SetMACLocalMACAddr(pMACBuffer);

		bRet = InitializePcap(strParam);
		if( FALSE == bRet )
		{
			DbgPrint("    [PktSender] Fail to initialize pcap");
		}


		/* initialize device */
		bRet = InitializePktSender();
		if( FALSE == bRet )
		{
			DbgPrint("    [PktSender] Fail to initialize pkt sender");
		}

	}
	else
	{
		DbgPrint("    [PktSender] Active IF address is NULL");
	}

	m_ulCntPktSendFail = 0;
	m_ulCntTotPktSend  = 0; 
	m_ulCnt2ndPktSendFail= 0;
}


CPktSender::~CPktSender(void)
{
}

BOOL CPktSender::InitializePcap( string& strActiveIFAddr )
{
	pcap_if_t*    pDevList;
	pcap_if_t*    pDevIter;
	pcap_addr_t*  pAddress;
	char          errbuf[PCAP_ERRBUF_SIZE+1];
	char          szSource[] = {"rpcap://"};
	string        strIFAddress;
	
	
	pDevList = NULL;
	pDevIter = NULL;


	/* set active interface IP */
	//SetInterfaceIP(strActiveIFAddr);

	/* find all available interfaces and find active interface */
	if (pcap_findalldevs_ex(szSource, NULL, &pDevList, errbuf) == -1)
	{
		DbgPrint("    [PktSender] pcap_findalldevs_ex() fail: %s", errbuf );
		return FALSE;
	}


	 /* IP addresses */
	for( pDevIter = pDevList; pDevIter ; pDevIter = pDevIter->next )
	{
		//printf("\tAddress Family: #%d\n", pAddress->addr->sa_family);
		for( pAddress = pDevIter->addresses; pAddress ; pAddress = pAddress->next )
		{

			/* only for the not null item */
			if( NULL != pAddress->addr )
			{
				strIFAddress = inet_ntoa( ((struct sockaddr_in *)pAddress->addr)->sin_addr);

				//if( 0 == strIFAddress.compare(m_strUsrSetInterfaceIP) )
				if( 0 == strIFAddress.compare(strActiveIFAddr) )
				{
					DbgPrint("[PktSender] IF selection info");
					DbgPrint("[PktSender] Name: %s", pDevIter->name);
					DbgPrint("[PktSender] Description: %s", pDevIter->description);
					
					/* copy interface name */
					m_strActiveIF = (pDevIter->name) + sizeof(szSource) - 1;
					
					pcap_freealldevs(pDevList);

					return TRUE;
				}

			}
		} // end of inner for loop

	}// end of for loop

	/* dealloc memory */
	pcap_freealldevs(pDevList);

	return FALSE;
}


void CPktSender::SetMACLocalMACAddr( unsigned char* pMAC )
{
	if( NULL != pMAC )
	{
		memcpy(&m_aucLocalMAC[0], pMAC, sizeof(m_aucLocalMAC) );
	}
}

BOOL CPktSender::InitializePktSender()
{
	m_pDevPcapHndl = pcap_open(m_strActiveIF.c_str(),      // name of the device
                               1700,                        // portion of the packet to capture (only the first 100 bytes)
                               PCAP_OPENFLAG_MAX_RESPONSIVENESS | PCAP_OPENFLAG_NOCAPTURE_LOCAL,//PCAP_OPENFLAG_PROMISCUOUS,  // promiscuous mode
                               1000,                       // read timeout
                               NULL,                       // authentication on the remote machine
                               m_achLastError              // error buffer
                               );

    /* Open the output device */
	if ( NULL == m_pDevPcapHndl)
    {
        //fprintf(stderr,"\nUnable to open the adapter. %s is not supported by WinPcap\n", argv[1]);
		//cout<<"[PktSender] Unable to open the adapter"<<endl;
		DbgPrint("    [PktSender] Unable to open the adapter");
        return FALSE;
    }

	return TRUE;
}

BOOL CPktSender::SendPkt( unsigned char* pucBuffer, unsigned int ulLength )
{
	unsigned char*  pucPktBuffer;
	unsigned int    ulPktLength;
#if 0
	WORD            wCheckSum;
	WORD            wnsCheckSum;
#endif

	
	ulPktLength  = ulLength + 14;
	pucPktBuffer = new unsigned char[ulPktLength];  // ethernet header + data length
	
	if( NULL == pucPktBuffer )
	{
		//cout<<"[PktSender] Fail to allocate send packet buffer"<<endl;
		DbgPrint("    [PktSender] Fail to allocate send packet buffer");
		return FALSE;
	}
	else
	{
#ifdef MEM_DEBUG
		char szMessage[1024];
		ZeroMemory( szMessage, sizeof(szMessage) );
		int nSize = sprintf_s(szMessage, "0x%08x@File: %s, Line:%d", pucPktBuffer, __FILE__, __LINE__);
		Mem_Debug(szMessage, nSize, TRUE );
#endif
	}

	/* initialize memroy */
	ZeroMemory( pucPktBuffer, ulPktLength);

	/* Supposing to be on ethernet, set mac destination to 1:1:1:1:1:1 */
	/*pucPktBuffer[0] = m_aucLocalMAC[0];
    pucPktBuffer[1] = m_achLocalMAC[1];
    pucPktBuffer[2] = m_achLocalMAC[2];
    pucPktBuffer[3] = m_achLocalMAC[3];
    pucPktBuffer[4] = m_achLocalMAC[4];
    pucPktBuffer[5] = m_achLocalMAC[5];*/
	
	memcpy(&pucPktBuffer[0], &m_aucLocalMAC[0], 6 );
    
    /* set mac source to 2:2:2:2:2:2 */
    pucPktBuffer[6]  = 0x02;
    pucPktBuffer[7]  = 0x02;
    pucPktBuffer[8]  = 0x02;
    pucPktBuffer[9]  = 0x02;
    pucPktBuffer[10] = 0x02;
    pucPktBuffer[11] = 0x02;
	//memset(&pucPktBuffer[6], 0x01, 6 );

	/* type = IP */
	pucPktBuffer[12] = 0x08;
	pucPktBuffer[13] = 0x00;
    
    /* copy IP pkt to generated packet */
	memcpy( &pucPktBuffer[14], pucBuffer, ulLength );


    /* Send down the packet */
	//m_ulCntTotPktSend++;
    if(pcap_sendpacket(m_pDevPcapHndl, pucPktBuffer, ulPktLength) != 0)
    {
        //printf("\nError sending the packet: %s", pcap_geterr(m_pDevPcapHndl));
		//cout<<"[PktSender] Fail to send packet - " << m_achLastError <<endl;
		//printf("\n[PktSender] Fail to send packet");
		//m_ulCntPktSendFail++;
		//printf("\n[PktSender] Fail : %d", m_ulCntPktSendFail);
		//printf("\n[PktSender] Try  : %d", m_ulCntTotPktSend);
		/*if(pcap_sendpacket(m_pDevPcapHndl, pucPktBuffer, ulPktLength) != 0)
		{
			m_ulCnt2ndPktSendFail++;
			printf("\n[PktSender] 2nd Fail: %d", m_ulCnt2ndPktSendFail );

		}*/
		/*
		MyPkt pkt;
		pkt.pPktData = pucPktBuffer;
		pkt.ulSize   = ulPktLength;
		m_listRetryPkt.push_back(pkt);
		return FALSE;
		*/


#ifdef MEM_DEBUG
		char szMessage[1024];
		ZeroMemory( szMessage, sizeof(szMessage) );
		int nSize = sprintf_s(szMessage, "0x%08x@File: %s, Line:%d", pucPktBuffer, __FILE__, __LINE__);
		Mem_Debug(szMessage, nSize, FALSE );
#endif
		delete [] pucPktBuffer;
        return FALSE;
    }
	
#ifdef MEM_DEBUG
		char szMessage[1024];
		ZeroMemory( szMessage, sizeof(szMessage) );
		int nSize = sprintf_s(szMessage, "0x%08x@File: %s, Line:%d", pucPktBuffer, __FILE__, __LINE__);
		Mem_Debug(szMessage, nSize, FALSE );
#endif
	delete [] pucPktBuffer;
    return TRUE;

}

