/*
* AMT Gateway Win32 version
*
* THE UNIVERSITY OF TEXAS AT DALLAS, ANDES Lab.
*
* Author: Juhyun Oh (Justin), justinoh@student.utdallas.edu
*
*/

#ifndef __COMMON_H__
#define __COMMON_H__

#define _WIN32_WINNT 0x0500
#include "pcap.h"
#include <Windows.h>
#include <process.h>
#include <iostream>
#include <list>
#include <ctime>
#include <string>
#include <fstream>
#include <vector>

//#include <Winternl.h>

using namespace std;

/* AMT protocol header type */
#define AMT_RELAY_DISCOVERY       (1)
#define AMT_RELAY_ADVERTISEMENT   (2)
#define AMT_REQUEST               (3)
#define AMT_MEMBERSHIP_QUERY      (4)
#define AMT_MEMBERSHIP_CHANGE     (5)
#define AMT_MCAST_DATA            (6)

#define AMT_PORT                  (2268)

#define DISCOVERY_TIMER_PERIOD		(20000)  // 20 sec
#define REQ_TIMER_INIT_PERIOD		(200000) // 200 sec, just for initiation
#define IGMPQUERY_TIMER_PERIOD		(60000)	 // 60 sec


/*
 *  relay status chage: 
 * 
 *  1.
 *  sys init(RELAY_NOT_FOUND) -> discovery start ( RELAY_DISCOVERY_INPROGRESS )->
 *  gte AMT_RELAY_ADVERTISEMENT message (RELAY_FOUND)
 * 
 *  2.  pending request trial excessdes its maximum -> discovery start ( RELAY_DISCOVERY_INPROGRESS )
 *
 */

#define RELAY_NOT_FOUND               (0)
#define RELAY_FOUND                   (1)
#define RELAY_DISCOVERY_INPROGRESS    (2)



/*
*  message related 
*/
#define RESPONSE_MAC_LEN              (6)
#define RESPONSE_NONCE_LEN            (4)
#define RESPONSE_RELAYIP_LEN          (4)
#define RESPONSE_REQMSG_LEN           (12)
#define RESPONSE_MEMBERSHIP_HADER_LEN (12)
#define GW_REQUEST_MAX                (3)//(7)
#define GW_REQUEST_RETRY_INTERVAL	  (1000)
#define UDP_SEND_RETRY                (3)

#define REQUEST_DELAY				  (2) // 2sec
#define REQUEST_SEND_INTERVAL		  (3) // 3sec

/* relay tunnel timeout */
//#define TUNNEL_TIMEOUT				  (480) // 8 min


/* pcap related */
#define MAX_PKT_CAPTURE_SIZE          (65536)

#define MALLOC(x) HeapAlloc(GetProcessHeap(), 0, (x))
#define FREE(x) HeapFree(GetProcessHeap(), 0, (x))

/* shared memory related */
#define PATH_BUF_SIZE 2048

/* player type */
#define WM_PLAYER	(0x01)
#define VLC_PLAYER	(0x02)
#define QT_PLAYER	(0x03)

/*
*  internal communcations
*/
#define SVR_PORT	(9124)

// command
#define MEMBERSHIP_UPDATE	(0x10)
//#define DROP_MEMEBRSHIP	(0x11)
#define PERIODIC_QUERY		(0x20)
#define MULTI_INSTANCE		(0x30)


inline void DbgPrint(const char* format,...)
{
#if 0
	return;
#else
	char szBuffer[1024];
	SYSTEMTIME stSysTime;

	ZeroMemory(szBuffer, sizeof(szBuffer));
	::GetLocalTime(&stSysTime);
	
	va_list args;
	va_start(args,format);
	vsprintf_s(szBuffer, format, args);
	va_end(args);
	printf("\n%04d-%02d-%02d %02d:%02d:%02d.%03d %s", stSysTime.wYear,
										stSysTime.wMonth,
										stSysTime.wDay,
										stSysTime.wHour,
										stSysTime.wMinute,
										stSysTime.wSecond,
										stSysTime.wMilliseconds,
										szBuffer);
#endif
}

//#define MEM_DEBUG

#ifdef MEM_DEBUG
extern HWND ghwndMemLogger;

inline void Mem_Debug( void* pMessage, int nLen, BOOL bAlloc )
{
	if( ghwndMemLogger != NULL )
	{
		COPYDATASTRUCT MyCDS;
		MyCDS.cbData = nLen;
		MyCDS.lpData = pMessage;

		if( bAlloc )
		{
			MyCDS.dwData = 1;
		}
		else
		{
			MyCDS.dwData = 0;
		}

		::SendMessage(ghwndMemLogger, WM_COPYDATA, 0, (LPARAM)(LPVOID)&MyCDS);
	}
}
#endif

/////////////////////////////////////////////////////////////////
//
//  struct definition  
//

typedef struct _PATH_INFO
{
	HANDLE hSMEvent;
	void*  pGroupInfoObj;
}PATH_INFO;


typedef struct _REQ_INFO
{
	void* pParent;              // pointer of parent
	unsigned int ulNonce;
}REQ_INFO;

/* struct for a request */
typedef struct _REQUEST
{
	HANDLE hReqTimerHndl;        /* request timer handle */
	unsigned int ulNonce;        /* nonce of this request */
	unsigned int ulReqCount;     /* request count */
	unsigned int ulIGMPLength;   /* IGMP packet length */
	unsigned char* pucIGMPData;
	REQ_INFO*    pReqInfo;
	//unsigned char ucIGMPData[IGMP_PKT_BUFFER_SIZE];   /* original IGMP packet buffer */

}REQUEST;

typedef struct _RELAY_MSG
{
	char* pcData;
	int   nDataLength;
}RELAY_MSG;


#endif